#  Authors: Sylvain Marie <sylvain.marie@se.com>
#
#  Copyright (c) Schneider Electric Industries, 2019. All right reserved.


class FieldTypeError(TypeError):  # FieldError
    """
    Error raised when the type of a field does not match expected type(s).
    """
    __slots__ = ('field', 'value', 'expected_types')

    def __init__(self, field, value, expected_types):
        self.field = field
        self.value = value
        # noinspection PyBroadException
        try:
            if len(expected_types) == 1:
                expected_types = expected_types[0]
        except:
            pass
        self.expected_types = expected_types

    def __str__(self):
        # representing the object might fail, protect ourselves
        # noinspection PyBroadException
        try:
            val_repr = repr(self.value)
        except Exception as e:
            val_repr = "<error while trying to represent value: %s>" % e

        # detail error message
        # noinspection PyBroadException
        try:
            # tuple or iterable of types ?
            sub_msg = "Value type should be one of (%s)" % ', '.join(("%s" % _t for _t in self.expected_types))
        except:  # noqa E722
            # single type
            sub_msg = "Value should be of type %s" % (self.expected_types,)

        return "Invalid value type provided for '%s'. %s. Instead, received a '%s': %s"\
               % (self.field.qualname, sub_msg, self.value.__class__.__name__, val_repr)


def _make_assert_is_of_type():
    try:
        from typeguard import check_type
        try:
            from typing import Union
        except ImportError:
            # (a) typing is not available, transform iterables of types into several calls
            def assert_is_of_type(field, value, typ):
                """
                Type checker relying on `typeguard` (python 3.5+)

                :param field:
                :param value:
                :param typ:
                :return:
                """
                try:
                    # iterate on the types
                    t_gen = (t for t in typ)
                except TypeError:
                    # not iterable : a single type
                    try:
                        check_type(field.qualname, value, typ)
                    except Exception as e:
                        # raise from
                        new_e = FieldTypeError(field, value, typ)
                        new_e.__cause__ = e
                        raise new_e
                else:
                    # iterate and try them all
                    e = None
                    for t in t_gen:
                        try:
                            check_type(field.qualname, value, typ)
                            return  # success !!!!
                        except Exception as e:
                            pass    # failed: lets try another one

                    # raise from
                    if e is not None:
                        new_e = FieldTypeError(field, value, typ)
                        new_e.__cause__ = e
                        raise new_e

        else:
            # (b) typing is available, use a Union
            def assert_is_of_type(field, value, typ):
                """
                Type checker relying on `typeguard` (python 3.5+)

                :param field:
                :param value:
                :param typ:
                :return:
                """
                try:
                    check_type(field.qualname, value, Union[typ])
                except Exception as e:
                    # raise from
                    new_e = FieldTypeError(field, value, typ)
                    new_e.__cause__ = e
                    raise new_e

    except ImportError:
        try:
            from pytypes import is_of_type

            def assert_is_of_type(field, value, typ):
                """
                Type checker relying on `pytypes` (python 2+)

                :param field:
                :param value:
                :param typ:
                :return:
                """
                try:
                    valid = is_of_type(value, typ)
                except Exception as e:
                    # raise from
                    new_e = FieldTypeError(field, value, typ)
                    new_e.__cause__ = e
                    raise new_e
                else:
                    if not valid:
                        raise FieldTypeError(field, value, typ)

        except ImportError:
            from valid8.utils.typing_inspect import is_typevar, is_union_type, get_args
            from valid8.utils.typing_tools import resolve_union_and_typevar

            def assert_is_of_type(field, value, typ):
                """
                Neither `typeguard` nor `pytypes` are available on this platform.

                This is a "light" implementation that basically resolves all `Union` and `TypeVar` into a flat list and
                then calls `isinstance`.

                :param field:
                :param value:
                :param typ:
                :return:
                """
                types = resolve_union_and_typevar(typ)
                try:
                    is_ok = isinstance(value, types)
                except TypeError as e:
                    if e.args[0].startswith("Subscripted generics cannot"):
                        raise TypeError("Neither typeguard not pytypes is installed - therefore it is not possible to "
                                        "validate subscripted typing structures such as %s" % types)
                    else:
                        raise
                else:
                    if not is_ok:
                        raise FieldTypeError(field, value, typ)

    return assert_is_of_type


try:  # very minimal way to check if typing it available, for runtime type checking
    # noinspection PyUnresolvedReferences
    from typing import Tuple
    assert_is_of_type = _make_assert_is_of_type()
except ImportError:
    assert_is_of_type = None



#  Authors: Sylvain Marie <sylvain.marie@se.com>
#
#  Copyright (c) Schneider Electric Industries, 2019. All right reserved.

from valid8 import Validator, ValidationError, ValidationFailure
from valid8.base import getfullargspec as v8_getfullargspec, get_callable_name, is_mini_lambda

from typing import Callable, Type, Any, TypeVar, Union, Iterable, Tuple, Mapping, Optional, Dict, Literal


T = TypeVar('T')

# ------------- validator type hints -----------
# 1. the lowest-level user or 3d party-provided validation functions
ValidationFunc = Union[Callable[[Any], Any],
                       Callable[[Any, Any], Any],
                       Callable[[Any, Any, Any], Any]]
"""A validation function is a callable with signature (val), (obj, val) or (obj, field, val), returning `True` 
or `None` in case of success"""

try:
    # noinspection PyUnresolvedReferences
    from mini_lambda import y
    ValidationFuncOrLambda = Union[ValidationFunc, type(y)]
except ImportError:
    ValidationFuncOrLambda = ValidationFunc

# 2. the syntax to optionally transform them into failure raisers by providing a tuple
ValidatorDef = Union[ValidationFuncOrLambda,
                     Tuple[ValidationFuncOrLambda, str],
                     Tuple[ValidationFuncOrLambda, Type[ValidationFailure]],
                     Tuple[ValidationFuncOrLambda, str, Type[ValidationFailure]]
                     ]
"""A validator is a validation function together with optional error message and error type"""

# 3. the syntax to describe several validation functions at once
VFDefinitionElement = Union[str, Type[ValidationFailure], ValidationFuncOrLambda]
"""This type represents one of the elements that can define a checker: help msg, failure type, callable"""

OneOrSeveralVFDefinitions = Union[ValidatorDef,
                                  Iterable[ValidatorDef],
                                  Mapping[VFDefinitionElement, Union[VFDefinitionElement,
                                                                     Tuple[VFDefinitionElement, ...]]]]
"""Several validators can be provided as a singleton, iterable, or dict-like. In that case the value can be a 
single variable or a tuple, and it will be combined with the key to form the validator. So you can use any of 
the elements defining a validators as the key."""

# shortcut name used everywhere. Less explicit
Validators = OneOrSeveralVFDefinitions


class FieldValidator(Validator):
    """
    Represents a `Validator` responsible to validate a `field`
    """
    __slots__ = '__weakref__', 'validated_field', 'base_validation_funcs'

    def __init__(self, validated_field: 'DescriptorField', validators: Validators,
                 error_type: 'Type[ValidationError]' = None, help_msg: str = None,
                 none_policy: int = None, **kw_context_args): ...

    def add_validator(self, validation_func: ValidatorDef): ...

    def get_callables_creator(self) -> Callable[[ValidationFunc, str, Type[ValidationFailure], ...],
                                                Callable[[Any, ...], type(None)]]: ...

    def get_additional_info_for_repr(self) -> str: ...

    def _get_name_for_errors(self, name: str) -> str: ...

    def assert_valid(self, obj: Any, value: Any, error_type: Type[ValidationError] = None,
                     help_msg: str = None, **ctx): ...


# --------------- converters
class Converter(object):
    __slots__ = ('name', )

    def __init__(self, name=None): ...

    def accepts(self, obj, field, value) -> Optional[bool]:
        ...

    def convert(self, obj, field, value) -> Any:
        ...

    @classmethod
    def create_from_fun(cls,
                        converter_fun: ConverterFuncOrLambda,
                        validation_fun: ValidationFuncOrLambda = None
                        ) -> Converter: ...

# noinspection PyAbstractClass
class ConverterWithFuncs(Converter):
    __slots__ = ('accepts', 'convert')

    def __init__(self, convert_fun, name=None, accepts_fun=None): ...

# --------------converter type hints
# 1. the lowest-level user or 3d party-provided validation functions
ConverterFunc = Union[Callable[[Any], Any],
                      Callable[[Any, Any], Any],
                      Callable[[Any, Any, Any], Any]]
"""A converter function is a callable with signature (val), (obj, val) or (obj, field, val), returning the 
converted value in case of success"""

try:
    # noinspection PyUnresolvedReferences
    from mini_lambda import y
    ConverterFuncOrLambda = Union[ConverterFunc, type(y)]
except ImportError:
    ConverterFuncOrLambda = ConverterFunc

# 2. the syntax to optionally transform them into Converter by providing a tuple
ValidType = Type
# noinspection PyUnboundLocalVariable
ConverterFuncDefinition = Union[Converter,
                                ConverterFuncOrLambda,
                                Tuple[ValidationFuncOrLambda, ConverterFuncOrLambda],
                                Tuple[ValidType, ConverterFuncOrLambda]]

TypeDef = Union[Type, Tuple[Type, ...], Literal['*'], str]  # todo remove str whe pycharm understands Literal
OneOrSeveralConverterDefinitions = Union[Converter,
                                         ConverterFuncOrLambda,
                                         Iterable[Tuple[TypeDef, ConverterFuncOrLambda]],
                                         Mapping[TypeDef, ConverterFuncOrLambda]]
Converters = OneOrSeveralConverterDefinitions


def make_3params_callable(f: Union[ValidationFunc, ConverterFunc],
                          is_mini_lambda: bool = False) -> Callable[[Any, 'Field', Any], Any]: ...


def make_converter(converter_def: ConverterFuncDefinition) -> Converter: ...

def make_converters_list(converters: OneOrSeveralConverterDefinitions) -> Tuple[Converter, ...]: ...

def trace_convert(field: 'Field', value: Any, obj: Any = None) -> Tuple[Any, DetailedConversionResults]: ...

class ConversionError(Exception):
    def __init__(self, value_to_convert, field, obj, err_dct): ...

def err_dct_to_str(err_dct: Dict[Converter, str]) -> str: ...

class DetailedConversionResults(object): ...



#  Authors: Sylvain Marie <sylvain.marie@se.com>
#
#  Copyright (c) Schneider Electric Industries, 2019. All right reserved.
import sys

import pytest

from pyfields import autofields, field, FieldTypeError, Field, get_fields
from pyfields.core import NativeField


@pytest.mark.parametrize("with_type_hints,type_check", [(False, False), (True, False), (True, True)])
def test_autofields_basic(with_type_hints, type_check):
    """tests that basic functionality of @autofields is ok """

    if with_type_hints:
        if sys.version_info < (3, 6):
            pytest.skip("Type annotations are not supported in python < 3.6")

        from ._test_py36 import _test_autofields
        Foo = _test_autofields(type_check)

        # test it
        assert isinstance(Foo.__dict__['barcls'], NativeField)
        assert isinstance(Foo.__dict__['barfunc'], Field)
        assert not isinstance(Foo.__dict__['fct'], Field)
        assert not isinstance(Foo.__dict__['cls'], Field)

        f = Foo(foo=1, barbar='yo', barfunc=lambda x: 2, barcls=str)
        if type_check:
            with pytest.raises(FieldTypeError):
                f.foo = 'ha'
        else:
            f.foo = 'ha'

        assert f.bar == 0
        assert f.fct() == 1
        assert f.barfunc(1) == 2
        assert f.barcls == str

    else:
        # retrocompatbility mode for python < 3.6
        # note: we also use this opportunity to test with parenthesis
        @autofields(check_types=type_check)
        class Foo(object):
            CONSTANT = 's'
            __a__ = 0

            foo = field()
            bar = 0     # type: int
            barcls = float
            barfunc = lambda x: x
            barbar = 0  # type: str

            class cls:
                pass

            def fct(self):
                return 1

        # test it
        assert isinstance(Foo.__dict__['barcls'], Field)
        assert isinstance(Foo.__dict__['barfunc'], Field)
        assert not isinstance(Foo.__dict__['fct'], Field)
        assert not isinstance(Foo.__dict__['cls'], Field)

        f = Foo(foo=1, barfunc=lambda x: 2, barcls=str)
        assert f.bar == 0
        assert f.fct() == 1
        assert f.barfunc(1) == 2
        assert f.barcls == str


def test_autofields_property_descriptors():
    """Checks that properties and descriptors are correctly ignored by autofields"""

    @autofields
    class Foo(object):
        foo = 1
        @property
        def bar(self):
            return 2

        class MyDesc():
            def __get__(self):
                return 1

        class MyDesc2():
            def __get__(self):
                return 0
            def __set__(self, instance, value):
                return

        m = MyDesc()
        p = MyDesc2()

    fields = get_fields(Foo)
    assert len(fields) == 1
    assert fields[0].name == 'foo'



#  Authors: Sylvain Marie <sylvain.marie@se.com>
#
#  Copyright (c) Schneider Electric Industries, 2019. All right reserved.

import sys
from enum import Enum
from textwrap import dedent
from inspect import getmro

try:
    from inspect import signature, Parameter
except ImportError:
    # noinspection PyUnresolvedReferences,PyPackageRequirements
    from funcsigs import signature, Parameter

from valid8 import ValidationFailure, is_pep484_nonable

from pyfields.typing_utils import assert_is_of_type, FieldTypeError
from pyfields.validate_n_convert import FieldValidator, make_converters_list, trace_convert

try:  # python 3.5+
    # noinspection PyUnresolvedReferences
    from typing import List, Callable, Type, Any, Union, Iterable, Tuple, TypeVar
    _NoneType = type(None)
    use_type_hints = sys.version_info > (3, 0)
    if use_type_hints:
        T = TypeVar('T')
        # noinspection PyUnresolvedReferences
        from pyfields.validate_n_convert import ValidatorDef, Validators, Converters, ConverterFuncDefinition,\
            DetailedConversionResults, ValidationFuncOrLambda, ValidType

except ImportError:
    use_type_hints = False


USE_ADVANCED_TYPE_CHECKER = assert_is_of_type is not None


PY36 = sys.version_info >= (3, 6)
get_type_hints = None
if PY36:
    try:
        from typing import get_type_hints
    except ImportError:
        pass

PY2 = sys.version_info < (3, 0)
# PY35 = sys.version_info >= (3, 5)

try:
    object.__qualname__
except AttributeError:
    # old python without __qualname__
    import re
    RE_CLASS_NAME = re.compile("<class '(.*)'>")

    def qualname(cls):
        cls_str = str(cls)
        match = RE_CLASS_NAME.match(cls_str)
        if match:
            return match.groups()[0]
        else:
            return cls_str


class FieldError(Exception):
    """
    Base class for exceptions related to fields
    """
    pass


class MandatoryFieldInitError(FieldError):
    """
    Raised by `field` when a mandatory field is read without being set first.
    """
    __slots__ = 'field_name', 'obj'

    def __init__(self, field_name, obj):
        self.field_name = field_name
        self.obj = obj

    def __str__(self):
        return "Mandatory field '%s' has not been initialized yet on instance %s." % (self.field_name, self.obj)


class ReadOnlyFieldError(FieldError):
    """
    Raised by descriptor field when a read-only attribute is accessed for writing
    """
    __slots__ = 'field_name', 'obj'

    def __init__(self, field_name, obj):
        self.field_name = field_name
        self.obj = obj

    def __str__(self):
        return "Read-only field '%s' has already been initialized on instance %s and cannot be modified anymore." \
               % (self.field_name, self.obj)


class Symbols(Enum):
    """
    A few symbols used in `fields` for signatures

    note: we used to use the great `sentinel` package to create these symbols one by one, but since we have
    now quite a number of symbols, it seemed overkill to create one anonymous class for each.

    Still, I am not sure if this made a perf difference actually.
    """
    GUESS = 0
    UNKNOWN = 1
    EMPTY = 2  # type: Any
    USE_FACTORY = 3
    _unset = 4

    def __repr__(self):
        """ More compact representation for signatures readability"""
        return self.name


# GUESS = sentinel.create('guess')
GUESS = Symbols.GUESS

# UNKNOWN = sentinel.create('unknown')
UNKNOWN = Symbols.UNKNOWN

# EMPTY = sentinel.create('empty')
EMPTY = Symbols.EMPTY

# USE_FACTORY = sentinel.create('use_factory')
USE_FACTORY = Symbols.USE_FACTORY

# _unset = sentinel.create('_unset')
_unset = Symbols._unset


if not PY36:
    # a thread-safe lock for the global instance counter
    from threading import Lock
    threadLock = Lock()


class Field(object):
    """
    Base class for fields
    """
    __slots__ = ('__weakref__', 'is_mandatory', 'default', 'is_default_factory', 'name', 'type_hint', 'nonable', 'doc',
                 'owner_cls', 'pending_validators', 'pending_converters')
    if not PY36:
        # we need to count the instances created, so as to be able to track their order in classes
        # indeed in python < 3.6, class members are not sorted by order of appearance.
        __slots__ += ('__fieldinstcount__', )
        __field_global_inst_counter__ = 0

    def __init__(self,
                 default=EMPTY,         # type: T
                 default_factory=None,  # type: Callable[[], T]
                 type_hint=EMPTY,       # type: Any
                 nonable=UNKNOWN,       # type: Union[bool, Symbols]
                 doc=None,              # type: str
                 name=None              # type: str
                 ):
        """See help(field) for details"""

        if not PY36:
            with threadLock:
                # remember the instance creation number, and increment the counter
                self.__fieldinstcount__ = Field.__field_global_inst_counter__
                Field.__field_global_inst_counter__ += 1

        # default
        if default_factory is not None:
            self.is_mandatory = False
            if default is not EMPTY:
                raise ValueError("Only one of `default` and `default_factory` should be provided")
            else:
                self.default = default_factory
                self.is_default_factory = True
        else:
            self.is_mandatory = default is EMPTY
            self.default = default
            self.is_default_factory = False

        # name
        self.name = name
        self.owner_cls = None

        # doc
        self.doc = dedent(doc) if doc is not None else None

        # type hints
        if type_hint is not EMPTY and type_hint is not None:
            self.type_hint = type_hint
        else:
            self.type_hint = EMPTY

        # nonable
        if nonable is GUESS:
            if self.default is None:
                self.nonable = True
            elif type_hint is not EMPTY and type_hint is not None:
                if is_pep484_nonable(type_hint):
                    self.nonable = True
                else:
                    self.nonable = UNKNOWN
            else:
                # set as unknown until type hint is set (in set_as_cls_member)
                self.nonable = UNKNOWN
        else:
            self.nonable = nonable

        # pending validators and converters
        self.pending_validators = None
        self.pending_converters = None

    def set_as_cls_member(self,
                          owner_cls,
                          name,
                          owner_cls_type_hints=None,
                          type_hint=None
                          ):
        """
        Updates a field with all information available concerning how it is attached to the class.

         - its owner class
         - the name under which it is known in that class
         - the type hints (python 3.6)

        In python 3.6+ this is called directly at class creation time through the `__set_name__` callback.

        In older python versions this is called whenever we have the opportunity :(, through `collect_fields`,
        `fix_fields` and `fix_field`. We currently use the following strategies in python 2 and 3.5-:

         - When users create a init method, `collect_fields` will be called when the init method is first accessed
         - When users GET a native field, or GET or SET a descriptor field, `fix_field` will be called.

        :param owner_cls:
        :param name:
        :param owner_cls_type_hints:
        :param type_hint: you can provide the type hint directly
        :return:
        """
        # set the owner class
        self.owner_cls = owner_cls

        if PY2 and isinstance(self, DescriptorField) and not issubclass(owner_cls, object):
            raise ValueError("descriptor fields can not be used on old-style classes under python 2.")

        # check if the name provided as argument differ from the one provided
        if self.name is not None:
            if self.name != name:
                raise ValueError("field name '%s' in class '%s' does not correspond to explicitly declared name '%s' "
                                 "in field constructor" % (name, owner_cls, self.name))
            # already set correctly
        else:
            # set it
            self.name = name

        # if not already manually overridden, get the type hints if there are some in the owner class annotations
        if self.type_hint is EMPTY:
            if owner_cls_type_hints is not None:
                if type_hint is not None:
                    raise ValueError("Provide either owner_cls_type_hints or type_hint, not both")
                type_hint = owner_cls_type_hints.get(name)

            if type_hint is not None:
                # only use type hint if not empty
                self.type_hint = type_hint
                # update the 'nonable' status
                if self.nonable is UNKNOWN:
                    if is_pep484_nonable(type_hint):
                        self.nonable = True
                    else:
                        self.nonable = UNKNOWN

        # detect a validator or a converter on a native field
        if self.pending_validators is not None or self.pending_converters is not None:
            # create a descriptor field to replace this native field
            new_field = DescriptorField.create_from_field(self, validators=self.pending_validators,
                                                          converters=self.pending_converters)
            # register it on the class in place of self
            setattr(self.owner_cls, self.name, new_field)

        # detect classes with slots
        elif not isinstance(self, DescriptorField) and '__slots__' in vars(owner_cls) \
                and '__dict__' not in owner_cls.__slots__:
            # create a descriptor field to replace of this native field
            new_field = DescriptorField.create_from_field(self)
            # register it on the class in place of self
            setattr(owner_cls, name, new_field)

    def __set_name__(self,
                     owner,  # type: Type[Any]
                     name    # type: str
                     ):
        if owner is not None:
            # fill all the information about how it is attached to the class
            cls_type_hints = get_type_hints(owner)
            self.set_as_cls_member(owner, name, owner_cls_type_hints=cls_type_hints)

    @property
    def qualname(self):
        # type: (...) -> str

        if self.owner_cls is not None:
            try:
                owner_qualname = self.owner_cls.__qualname__
            except AttributeError:
                # python 2: no __qualname__
                owner_qualname = qualname(self.owner_cls)
        else:
            owner_qualname = "<unknown_cls>"

        return "%s.%s" % (owner_qualname, self.name)

    def __repr__(self):
        return "<%s: %s>" % (self.__class__.__name__, self.qualname)

    def default_factory(self, f):
        """
        Decorator to register the decorated function as the default factory of a field. Any previously registered
        default factory will be overridden.

        The decorated function should accept a single argument `(obj/self)`, and should return a value to use as the
        default.

        >>> import sys, pytest
        >>> if sys.version_info < (3, 6): pytest.skip("doctest skipped for python < 3.6")
        ...
        >>> class Pocket:
        ...     items = field()
        ...
        ...     @items.default_factory
        ...     def default_items(self):
        ...         print("generating default value for %s" % self)
        ...         return []
        ...
        >>> p = Pocket()
        >>> p.items
        generating default value for <pyfields.core.Pocket object ...
        []

        """
        self.default = f
        self.is_default_factory = True
        self.is_mandatory = False
        return f

    def validator(self,
                  help_msg=None,     # type: str
                  failure_type=None  # type: Type[ValidationFailure]
                  ):
        """
        A decorator to add a validator to a field.

        >>> import sys, pytest
        >>> if sys.version_info < (3, 6): pytest.skip('skipped on python <3.6')
        ...
        >>> class Foo(object):
        ...     m = field()
        ...     @m.validator
        ...     def m_is_positive(self, m_value):
        ...         return m_value > 0
        ...
        >>> o = Foo()
        >>> o.m = 0  # doctest: +NORMALIZE_WHITESPACE
        Traceback (most recent call last):
        ...
        valid8.entry_points.ValidationError[ValueError]: Error validating [Foo.m=0]. InvalidValue:
            Function [m_is_positive] returned [False] for value 0.

        The decorated function should have a signature of `(val)`, `(obj/self, val)`, or `(obj/self, field, val)`. It
        should return `True` or `None` in case of success.

        You can use several of these decorators on the same function so as to share implementation across multiple
        fields:

        >>> class Foo(object):
        ...     m = field()
        ...     m2 = field()
        ...
        ...     @m.validator
        ...     @m2.validator
        ...     def is_positive(self, field, value):
        ...         print("validating %s" % field.qualname)
        ...         return value > 0
        ...
        >>> o = Foo()
        >>> o.m2 = 12
        validating Foo.m2
        >>> o.m = 0  # doctest: +NORMALIZE_WHITESPACE
        Traceback (most recent call last):
        ...
        valid8.entry_points.ValidationError[ValueError]: Error validating [Foo.m=0]. InvalidValue:
            Function [is_positive] returned [False] for value 0.

        :param help_msg:
        :param failure_type:
        :return:
        """
        if help_msg is not None and callable(help_msg) and failure_type is None:
            # used without parenthesis @<field>.validator:  validation_callable := help_msg
            self.add_validator(help_msg)
            return help_msg
        else:
            # used with parenthesis @<field>.validator(...): return a decorator
            def decorate_f(f):
                # create the validator definition
                if help_msg is None:
                    if failure_type is None:
                        validator = f
                    else:
                        validator = (f, failure_type)
                else:
                    if failure_type is None:
                        validator = (f, help_msg)
                    else:
                        validator = (f, help_msg, failure_type)
                self.add_validator(validator)
                return f

            return decorate_f

    def add_validator(self,
                      validator  # type: ValidatorDef
                      ):
        """
        Adds a validator to the set of validators on that field.
        This is the implementation for native fields

        :param validator:
        :return:
        """
        if self.owner_cls is not None:
            # create a descriptor field instead of this native field
            new_field = DescriptorField.create_from_field(self, validators=(validator, ))

            # register it on the class
            setattr(self.owner_cls, self.name, new_field)
        else:
            if not PY36:
                raise UnsupportedOnNativeFieldError(
                    "defining validators is not supported on native fields in python < 3.6."
                    " Please set `native=False` on field '%s' to enable this feature."
                    % (self,))

            # mark as pending
            if self.pending_validators is None:
                self.pending_validators = [validator]
            else:
                self.pending_validators.append(validator)

    def converter(self,
                  _decorated_fun=None,  # type: _NoneType
                  accepts=None,         # type: Union[ValidationFuncOrLambda, ValidType]
                  ):
        """
        A decorator to add a validator to a field.

        >>> import sys, pytest
        >>> if sys.version_info < (3, 6): pytest.skip('skipped on python <3.6')
        ...
        >>> class Foo(object):
        ...     m = field()
        ...     @m.converter
        ...     def m_from_anything(self, m_value):
        ...         return int(m_value)
        ...
        >>> o = Foo()
        >>> o.m = '0'
        >>> o.m
        0

        The decorated function should have a signature of `(val)`, `(obj/self, val)`, or `(obj/self, field, val)`. It
        should return a converted value in case of success.

        You can explicitly declare which values are accepted by the converter, by providing an `accepts` argument.
        It may either contain a `<validation_callable>`, an `<accepted_type>` or a wildcard (`'*'` or `None`). Passing
        a wildcard is equivalent to calling the decorator without parenthesis as seen above.
        WARNING: this argument needs to be provided as keyword for the converter to work properly.

        You can use several of these decorators on the same function so as to share implementation across multiple
        fields:

        >>> class Foo(object):
        ...     m = field(type_hint=int, check_type=True)
        ...     m2 = field(type_hint=int, check_type=True)
        ...
        ...     @m.converter(accepts=str)
        ...     @m2.converter
        ...     def from_anything(self, field, value):
        ...         print("converting a value for %s" % field.qualname)
        ...         return int(value)
        ...
        >>> o = Foo()
        >>> o.m2 = '12'
        converting a value for Foo.m2
        >>> o.m2 = 1.5
        converting a value for Foo.m2
        >>> o.m = 1.5  # doctest: +NORMALIZE_WHITESPACE
        Traceback (most recent call last):
        ...
        pyfields.typing_utils.FieldTypeError: Invalid value type provided for 'Foo.m'. Value should be of type
          <class 'int'>. Instead, received a 'float': 1.5

        :param _decorated_fun: internal, the decorated function. Do not fill this argument!
        :param accepts: a `<validation_callable>`, an `<accepted_type>` or a wildcard (`'*'` or `None`) defining on
            which values this converter will have a chance to succeed. Default is `None`.
        :return:
        """
        if accepts is None and _decorated_fun is not None:
            # used without parenthesis @<field>.converter:
            self.add_converter(_decorated_fun)
            return _decorated_fun
        else:
            # used with parenthesis @<field>.converter(...): return a decorator
            def decorate_f(f):
                # create the converter definition
                self.add_converter((accepts, f))
                return f

            return decorate_f

    def add_converter(self,
                      converter_def  # type: ConverterFuncDefinition
                      ):
        """
        Adds a converter to the set of converters on that field.
        This is the implementation for native fields.

        :param converter_def:
        :return:
        """
        if self.owner_cls is not None:
            # create a descriptor field instead of this native field
            new_field = DescriptorField.create_from_field(self, converters=(converter_def, ))

            # register it on the class as a replacement for this native field
            setattr(self.owner_cls, self.name, new_field)
        else:
            if not PY36:
                raise UnsupportedOnNativeFieldError(
                    "defining converters is not supported on native fields in python < 3.6."
                    " Please set `native=False` on field '%s' to enable this feature."
                    % (self,))

            # mark as pending
            if self.pending_converters is None:
                self.pending_converters = [converter_def]
            else:
                self.pending_converters.append(converter_def)

    def trace_convert(self, value, obj=None):
        # type: (...) -> Tuple[Any, DetailedConversionResults]
        """
        Can be used to debug conversion problems.
        Instead of just returning the converted value, it also returns conversion details.

        Note that this method does not set the field value, it simply returns the conversion results.
        In case no converter is able to convert the provided value, a `ConversionError` is raised.

        :param obj:
        :param value:
        :return: a tuple (converted_value, details).
        """
        raise UnsupportedOnNativeFieldError("Native fields do not have converters.")


def field(type_hint=None,        # type: Union[Type[T], Iterable[Type[T]]]
          nonable=GUESS,         # type: Union[bool, Type[GUESS]]
          check_type=False,      # type: bool
          default=EMPTY,         # type: T
          default_factory=None,  # type: Callable[[], T]
          validators=None,       # type: Validators
          converters=None,       # type: Converters
          read_only=False,       # type: bool
          doc=None,              # type: str
          name=None,             # type: str
          native=None            # type: bool
          ):
    # type: (...) -> Union[T, Field]
    """
    Returns a class-level attribute definition. It allows developers to define an attribute without writing an
    `__init__` method. Typically useful for mixin classes.

    Laziness
    --------
    The field will be lazily-defined, so if you create an instance of the class, the field will not have any value
    until it is first read or written.

    Optional/Mandatory
    ------------------
    By default fields are mandatory, which means that you must set them before reading them (otherwise a
    `MandatoryFieldInitError` will be raised). You can define an optional field by providing a `default` value.
    This value will not be copied but used "as is" on all instances, following python's classical pattern for default
    values. If you wish to run specific code to instantiate the default value, you may provide a `default_factory`
    callable instead. That callable should have no mandatory argument and should return the default value. Alternately
    you can use the `@<field>.default_factory` decorator.

    Read-only
    ---------
    TODO

    Typing
    ------
    Type hints for fields can be provided using the standard python typing mechanisms (type comments for python < 3.6
    and class member type hints for python >= 3.6). Types declared this way will not be checked at runtime, they are
    just hints for the IDE. You can also specify a `type_hint` explicitly to override the type hints gathered from the
    other means indicated above. It supports both a single type or an iterable of alternate types (e.g. `(int, str)`).
    The corresponding type hint is automatically declared by `field` so your IDE will know about it. Specifying a
    `type_hint` explicitly is mostly useful if you are running python < 3.6 and wish to use type validation, see below.

    By default `check_type` is `False`. This means that the above mentioned `type_hint` is just a hint. If you set
    `check_type=True` the type declared in the type hint will be validated, and a `FieldTypeError` will be raised if
    provided values are invalid. Important: if you are running python < 3.6 you have to set the type hint explicitly
    using `type_hint` if you wish to set `check_type=True`, otherwise you will get an exception. Indeed type comments
    can not be collected by the code.

    Type hints relying on the `typing` module (PEP484) are correctly checked using whatever 3d party type checking
    library is available (`typeguard` is first looked for, then `pytypes` as a fallback). If none of these providers
    are available, a fallback implementation is provided, basically flattening `Union`s and replacing `TypeVar`s before
    doing `is_instance`. It is not guaranteed to support all `typing` subtleties.

    Validation
    ----------
    TODO

    Nonable
    -------
    TODO
    see also: https://stackoverflow.com/a/57390124/7262247

    Conversion
    ----------
    TODO

    Documentation
    -------------
    A docstring can be provided for code readability.

    Example
    -------

    >>> import sys, pytest
    >>> if sys.version_info < (3, 6): pytest.skip('skipped on python <3.6')
    ...
    >>> class Foo(object):
    ...     od = field(default='bar', doc="This is an optional field with a default value")
    ...     odf = field(default_factory=lambda obj: [], doc="This is an optional with a default value factory")
    ...     m = field(doc="This is a mandatory field")
    ...     mt: int = field(check_type=True, doc="This is a type-checked mandatory field")
    ...
    >>> o = Foo()
    >>> o.od   # read access with default value
    'bar'
    >>> o.odf  # read access with default value factory
    []
    >>> o.odf = 12  # write access
    >>> o.odf
    12
    >>> o.m  # read access for mandatory attr without init
    Traceback (most recent call last):
        ...
    pyfields.core.MandatoryFieldInitError: Mandatory field 'm' has not been initialized yet on instance...
    >>> o.m = True
    >>> o.m  # read access for mandatory attr after init
    True
    >>> del o.m  # all attributes can be deleted, same behaviour than new object
    >>> o.m
    Traceback (most recent call last):
        ...
    pyfields.core.MandatoryFieldInitError: Mandatory field 'm' has not been initialized yet on instance...
    >>> o.mt = 1
    >>> o.mt = '1'
    Traceback (most recent call last):
        ...
    pyfields.typing_utils.FieldTypeError: Invalid value type ...

    Limitations
    -----------
    Old-style classes are not supported: in python 2, don't forget to inherit from `object`.

    Performance overhead
    --------------------
    `field` has two different ways to create your fields. One named `NativeField` is faster but does not permit type
    checking, validation, or converters; besides it does not work with classes using `__slots__`. It is used by default
    everytime where it is possible, except if you use one of the above mentioned features. In that case a
    `DescriptorField` will transparently be created. You can force a `DescriptorField` to be created by setting
    `native=False`.

    The `NativeField` class implements the "non-data" descriptor protocol. So the first time the attribute is read, a
    small python method call extra cost is paid. *But* afterwards the attribute is replaced with a native attribute
    inside the object `__dict__`, so subsequent calls use native access without overhead.
    This was inspired by
    [werkzeug's @cached_property](https://tedboy.github.io/flask/generated/generated/werkzeug.cached_property.html).

    Inspired by
    -----------
    This method was inspired by

     - @lazy_attribute (sagemath)
     - @cached_property (werkzeug) and https://stackoverflow.com/questions/24704147/python-what-is-a-lazy-property
     - https://stackoverflow.com/questions/42023852/how-can-i-get-the-attribute-name-when-working-with-descriptor-protocol-in-python
     - attrs / dataclasses

    :param type_hint: an optional explicit type hint for the field, to override the type hint defined by PEP484
        especially on old python versions because type comments can not be captured. Both a single type or an iterable
        of alternate types (e.g. `(int, str)`) are supported. By default the type hint is just a
        hint and does not contribute to validation. To enable type validation, set `check_type` to `True`.
    :param nonable: a boolean that can be used to explicitly declare that a field can contain `None`. When this is set
        to an explicit `True` or `False` value, usual type checking and validation (*if any*) are not anymore executed
        on `None` values. Instead ; if this is `True`, type checking and validation will be *deactivated* when the field
        is set to `None` so as to always accept the value. If this is `False`, an `None`error will be raised when `None` is
        set on the field.
        When this is left as `GUESS` (default), the behaviour is "automatic". This means that
         - if the field (a) is optional with default value `None` or (b) has type hint `typing.Optional[]`, the
           behaviour will be the same as with `nonable=True`.
         - otherwise, the value will be the same as `nonable=UNKNOWN` and no special behaviour is put in place. `None`
           values will be treated as any other value. This can be particularly handy if a field accepts `None` ONLY IF
           another field is set to a special value. This can be done in a custom validator.
    :param check_type: by default (`check_type=False`), the type of a field, provided using PEP484 type hints or
        an explicit `type_hint`, is not validated when you assign a new value to it. You can activate type validation
        by setting `check_type=True`. In that case the field will become a descriptor field.
    :param default: a default value for the field. Providing a `default` makes the field "optional". `default` value
        is not copied on new instances, if you wish a new copy to be created you should provide a `default_factory`
        instead. Only one of `default` or `default_factory` should be provided.
    :param default_factory: a factory that will be called (without arguments) to get the default value for that
        field, everytime one is needed. Providing a `default_factory` makes the field "optional". Only one of `default`
        or `default_factory` should be provided.
    :param validators: a validation function definition, sequence of validation function definitions, or dict-like of
        validation function definitions. See `valid8` "simple syntax" for details
    :param converters: a sequence of (<type_def>, <converter>) pairs or a dict-like of such pairs. `<type_def>` should
        either be a type, a tuple of types, or the '*' string indicating "any other case".
    :param read_only: a boolean (default `False`) stating if a field can be modified after initial value has been
        provided.
    :param doc: documentation for the field. This is mostly for class readability purposes for now.
    :param name: in python < 3.6 this is mandatory if you do not use any other decorator or constructor creation on the
        class (such as `make_init`). If provided, it should be the same name than the one used used in the class field
        definition (i.e. you should define the field as `<name> = field(name=<name>)`).
    :param native: a boolean that can be turned to `False` to force a field to be a descriptor field, or to `True` to
        force it to be a native field. Native fields are faster but can not support type and value validation
        nor conversions or callbacks. `None` (default) automatically sets `native=True` if no `validators` nor
        `check_type=True` nor `converters` are provided ; and `native=False` otherwise. In general you should not
        set this option manually except for experiments.
    :return:
    """
    # Should we create a Native or a Descriptor field ?
    if native is None:
        # default: choose automatically according to user-provided options
        create_descriptor = check_type or (validators is not None) or (converters is not None) or read_only
    else:
        # explicit user choice
        if native:
            # explicit `native=True`.
            if check_type or (validators is not None) or (converters is not None) or read_only:
                raise UnsupportedOnNativeFieldError("`native=False` can not be set "
                                                    "if a `validators` or `converters` is provided "
                                                    "or if `check_type` or `read_only` is `True`")
            else:
                create_descriptor = False
        else:
            # explicit `native=False`. Force-use a descriptor
            create_descriptor = True

    # Create the correct type of field
    if create_descriptor:
        return DescriptorField(type_hint=type_hint, nonable=nonable, default=default, default_factory=default_factory,
                               check_type=check_type, validators=validators, converters=converters,
                               read_only=read_only, doc=doc, name=name)
    else:
        return NativeField(type_hint=type_hint, nonable=nonable, default=default, default_factory=default_factory,
                           doc=doc, name=name)


class UnsupportedOnNativeFieldError(FieldError):
    """
    Exception raised whenever someone tries to perform an operation that is not supported on a "native" field.
    """
    pass


class ClassFieldAccessError(FieldError):
    """
    Error raised when you use <cls>.<field>. This is currently put in place because otherwise the
    type hints in pycharm get messed up. See below.
    """
    __slots__ = 'field',

    # noinspection PyShadowingNames
    def __init__(self, field):
        self.field = field

    def __str__(self):
        return "Accessing a `field` from the class is not yet supported. You can use %s.__dict__['%s'] as a " \
               "workaround. See https://github.com/smarie/python-pyfields/issues/12" \
               % (self.field.owner_cls.__name__, self.field.name)


class NativeField(Field):
    """
    A field that is replaced with a native python attribute on first read or write access.
    Faster but provides not much flexibility (no validator, no type check, no converter)
    """
    __slots__ = ()

    def __get__(self, obj, obj_type):
        # type: (...) -> T

        # do this first, because a field might be referenced from its class the first time it will be used
        # for example if in `make_init` we use a field defined in another class, that was not yet accessed on instance.
        if not PY36 and self.name is None:
            # __set_name__ was not called yet. lazy-fix the name and type hints
            fix_field(obj_type, self)

        if obj is None:
            # class-level call ?
            # TODO put back when https://youtrack.jetbrains.com/issue/PY-38151 is solved
            # return self
            # even this does not work
            # exec("o = self", globals(), locals())
            # return locals()['o']
            raise ClassFieldAccessError(self)

        # Check if the field is already set in the object __dict__
        value = obj.__dict__.get(self.name, _unset)

        if value is _unset:
            # mandatory field: raise an error
            if self.is_mandatory:
                raise MandatoryFieldInitError(self.name, obj)

            # optional: get default
            if self.is_default_factory:
                value = self.default(obj)
            else:
                value = self.default

            # nominal initialization on first read: we set the attribute in the object __dict__
            # so that next reads will be pure native field access
            obj.__dict__[self.name] = value

        # else:
            # this was probably a manual call of __get__ (or a concurrent call of the first access)

        return value

    # not needed apparently
    # def __delete__(self, obj):
    #     try:
    #         del obj.__dict__[self.name]
    #     except KeyError:
    #         # silently ignore: the field has not been set on that object yet,
    #         # and we wont delete the class `field` anyway...
    #         pass


class NoneError(TypeError, ValueError, FieldError):
    """
    Error raised when `None` is set on an explicitly not-nonable field. It is both a `TypeError` and a `ValueError`.
    """
    __slots__ = ('field', )

    def __init__(self, field):
        super(NoneError, self).__init__()
        self.field = field

    def __str__(self):
        return "Received invalid value `None` for '%s'. This field is explicitely declared as non-nonable."\
               % (self.field.qualname, )


# default value policies
_NO = None
_NO_BUT_CAN_CACHE_FIRST_RESULT = False
_YES = True


class DescriptorField(Field):
    """
    General-purpose implementation for fields that require type-checking or validation or converter
    """
    __slots__ = 'root_validator', 'check_type', 'converters', 'read_only', '_default_is_safe'

    @classmethod
    def create_from_field(cls,
                          other_field,      # type: Field
                          validators=None,  # type: Iterable[ValidatorDef]
                          converters=None   # type: Iterable[ConverterFuncDefinition]
                          ):
        # type: (...) -> DescriptorField
        """
        Creates a descriptor field by copying the information from the given other field, typically a native field

        :param other_field:
        :param validators: validators to add to the field definition
        :param converters: converters to add to the field definition
        :return:
        """
        if other_field.is_default_factory:
            default = EMPTY
            default_factory = other_field.default
        else:
            default_factory = None
            default = other_field.default

        new_field = DescriptorField(type_hint=other_field.type_hint, default=default, default_factory=default_factory,
                                    doc=other_field.doc, name=other_field.name, validators=validators,
                                    converters=converters)

        # copy the owner class info too
        new_field.owner_cls = other_field.owner_cls
        return new_field

    def __init__(self,
                 type_hint=None,        # type: Type[T]
                 nonable=UNKNOWN,       # type: Union[bool, Symbols]
                 default=EMPTY,         # type: T
                 default_factory=None,  # type: Callable[[], T]
                 check_type=False,      # type: bool
                 validators=None,       # type: Validators
                 converters=None,       # type: Converters
                 read_only=False,       # type: bool
                 doc=None,              # type: str
                 name=None              # type: str
                 ):
        """See help(field) for details"""
        super(DescriptorField, self).__init__(type_hint=type_hint, nonable=nonable,
                                              default=default, default_factory=default_factory, doc=doc, name=name)

        # type validation
        self.check_type = check_type

        # validators
        if validators is not None:
            self.root_validator = FieldValidator(self, validators)
        else:
            self.root_validator = None

        # converters
        if converters is not None:
            self.converters = list(make_converters_list(converters))
        else:
            self.converters = None

        # read-only
        self.read_only = read_only

        # self._default_is_safe is used to know if we should validate/convert the default value before use
        #  - None means "always". This is the case when there is a default factory we can't modify
        #  - False means "once", and then True means "not anymore" (after first validation). This is the case
        #    when we can modify the default value so that we can replace it with the possibly converted one
        if default is not EMPTY:
            # a fixed default value is here, we'll validate it once and for all
            self._default_is_safe = _NO_BUT_CAN_CACHE_FIRST_RESULT
        elif default_factory is not None:
            # noinspection PyBroadException
            try:
                # is this the `copy_value` factory ?
                default_factory.clone_with_new_val
            except Exception:
                # No: the factory can be anything else
                self._default_is_safe = _NO
            else:
                # Yes: we can replace the value that it uses on first
                self._default_is_safe = _NO_BUT_CAN_CACHE_FIRST_RESULT
        else:
            # no default at all
            self._default_is_safe = _NO

    def add_validator(self,
                      validator  # type: ValidatorDef
                      ):
        """
        Add a validation function to the set of validation functions.

        :param validator:
        :return:
        """
        if self.root_validator is None:
            self.root_validator = FieldValidator(self, validator)
        else:
            self.root_validator.add_validator(validator)

    def add_converter(self,
                      converter_def  # type: ConverterFuncDefinition
                      ):
        converters = make_converters_list(converter_def)
        if self.converters is None:
            # use the new list
            self.converters = list(converters)
        else:
            # concatenate the lists
            self.converters += converters

    def __get__(self, obj, obj_type):
        # type: (...) -> T

        # do this first, because a field might be referenced from its class the first time it will be used
        # for example if in `make_init` we use a field defined in another class, that was not yet accessed on instance.
        if not PY36 and self.name is None:
            # __set_name__ was not called yet. lazy-fix the name and type hints
            fix_field(obj_type, self)

        if obj is None:
            # class-level call ?
            # TODO put back when https://youtrack.jetbrains.com/issue/PY-38151 is solved
            # return self
            raise ClassFieldAccessError(self)

        private_name = '_' + self.name

        # Check if the field is already set in the object
        value = getattr(obj, private_name, _unset)

        if value is _unset:
            # mandatory field: raise an error
            if self.is_mandatory:
                raise MandatoryFieldInitError(self.name, obj)

            # optional: get default
            if self.is_default_factory:
                value = self.default(obj)
            else:
                value = self.default

            # nominal initialization on first read: we set the attribute in the object
            if self._default_is_safe is _YES:
                # no need to validate/convert the default value, fast track (use the private name directly)
                setattr(obj, private_name, value)
            else:
                # we need conversion and validation - go through the setter (same as using the public name)
                possibly_converted_value = self.__set__(obj, value, _return=True)

                if self._default_is_safe is _NO_BUT_CAN_CACHE_FIRST_RESULT:
                    # there is a possibility to remember the new default and skip this next time

                    # If there was a conversion, use the converted value as the new default
                    if possibly_converted_value is not value:
                        if self.is_default_factory:
                            # Modify the `copy_value` factory
                            self.default = self.default.clone_with_new_val(possibly_converted_value)
                        else:
                            # Modify the value
                            self.default = possibly_converted_value
                    # else:
                    #     # no conversion: we can continue to use the same default value, it is valid
                    #     pass

                    # mark the default as safe now, so that this is skipped next time
                    self._default_is_safe = _YES

                return possibly_converted_value

        return value

    def trace_convert(self, value, obj=None):
        """Overrides the method in `Field` to provide a valid implementation."""
        return trace_convert(field=self, value=value, obj=obj)

    def __set__(self,
                obj,
                value,         # type: T
                _return=False  # type: bool
                ):

        # do this first, because a field might be referenced from its class the first time it will be used
        # for example if in `make_init` we use a field defined in another class, that was not yet accessed on instance.
        if not PY36 and self.name is None:
            # __set_name__ was not called yet. lazy-fix the name and type hints
            fix_field(obj.__class__, self)

        if obj is None:
            # class-level call ?
            # TODO put back when https://youtrack.jetbrains.com/issue/PY-38151 is solved
            # return self
            raise ClassFieldAccessError(self)

        if self.converters is not None:
            # this is an inlined version of `trace_convert` with no capture of details
            for converter in self.converters:
                # noinspection PyBroadException
                try:
                    # does the converter accept this input ?
                    accepted = converter.accepts(obj, self, value)
                except Exception:
                    # ignore all exceptions from converters
                    continue
                else:
                    if accepted is None or accepted:
                        # if so, let's try to convert
                        try:
                            converted_value = converter.convert(obj, self, value)
                        except Exception:
                            # ignore all exceptions from converters
                            continue
                        else:
                            # successful conversion: use the converted value
                            value = converted_value
                            break
                    else:
                        continue

        # speedup for vars used several time
        t = self.type_hint
        nonable = self.nonable
        private_name = "_" + self.name

        # read-only check
        if self.read_only:
            # Check if the field is already set in the object
            _v = getattr(obj, private_name, _unset)
            if _v is not _unset:
                raise ReadOnlyFieldError(self.qualname, obj)

        # type checker and validators
        if value is not None or nonable is UNKNOWN:
            # check the type
            if self.check_type:
                if t is EMPTY:
                    raise ValueError("`check_type` is enabled on field '%s' but no type hint is available. Please provide"
                                     "type hints or set `field.check_type` to `False`. Note that python code is not able to"
                                     " read type comments so if you wish to be compliant with python < 3.6 you'll have to"
                                     "set the type hint explicitly in `field.type_hint` instead")

                if USE_ADVANCED_TYPE_CHECKER:
                    # take into account all the subtleties from `typing` module by relying on 3d party providers.
                    assert_is_of_type(self, value, t)

                elif not isinstance(value, t):
                    raise FieldTypeError(self, value, t)

            # run the validators
            if self.root_validator is not None:
                self.root_validator.assert_valid(obj, value)

        elif not nonable:
            # value is None and field is not nonable: raise an error
            # note: the root validator might not even exist, so do not reuse valid8 none rejecter here
            raise NoneError(self)
        # else:
        #     # value is None and field is nonable: nothing to do
        #     pass

        # set the new value
        setattr(obj, private_name, value)

        # return it for the callers that need it
        if _return:
            return value

    def __delete__(self, obj):
        # private_name = "_" + self.name
        delattr(obj, "_" + self.name)


# noinspection PyShadowingNames
def fix_field(cls,                     # type: Type[Any]
              field,                   # type: Field
              include_inherited=True,  # type: bool
              fix_type_hints=PY36      # type: bool
              ):
    """
    Fixes the given field name and type hint on the given class

    :param cls:
    :param field:
    :param include_inherited: should the field be looked for in parent classes following the mro. Default = True
    :param fix_type_hints:
    :return:
    """
    if fix_type_hints:
        cls_type_hints = get_type_hints(cls)
    else:
        cls_type_hints = None

    where_cls = getmro(cls) if include_inherited else (cls, )

    found = False
    for _cls in where_cls:
        for member_name, member in vars(_cls).items():
            # if not member_name.startswith('__'):   not stated in the doc: too dangerous to have such implicit filter
            if member is field:
                # do the same than in __set_name__
                field.set_as_cls_member(_cls, member_name, owner_cls_type_hints=cls_type_hints)
                # found: no need to look further
                found = True
                break
        if found:
            break
    else:
        raise ValueError("field %s was not found on class %s%s"
                         % (field, cls, 'or its ancestors' if include_inherited else ''))



#  Authors: Sylvain Marie <sylvain.marie@se.com>
#
#  Copyright (c) Schneider Electric Industries, 2019. All right reserved.

import pytest

from typing import List
from pyfields import field, inject_fields, MandatoryFieldInitError, make_init, autofields


def _test_class_annotations():
    class Foo:
        field_with_validate_type: int = field(check_type=True)
        field_with_defaults: str = field()

    return Foo


def _test_readme_type_validation():
    class Wall(object):
        height: int = field(check_type=True, doc="Height of the wall in mm.")
        color: str = field(check_type=True, default='white', doc="Color of the wall.")

    return Wall


def _test_readme_value_validation(colors):
    from mini_lambda import x
    from valid8.validation_lib import is_in

    class Wall(object):
        height: int = field(validators={'should be a positive number': x > 0,
                                        'should be a multiple of 100': x % 100 == 0},
                            doc="Height of the wall in mm.")
        color: str = field(validators=is_in(colors),
                           default='white',
                           doc="Color of the wall.")

    return Wall


def test_value_validation_advanced(validate_width):
    class Wall(object):
        height: int = field(doc="Height of the wall in mm.")
        width: str = field(validators=validate_width,
                           doc="Width of the wall in mm.")
    return Wall


def _test_readme_constructor(explicit_fields_list, init_type, native):
    if init_type == 'inject_fields':
        if explicit_fields_list:
            class Wall:
                height: int = field(doc="Height of the wall in mm.", native=native)
                color: str = field(default='white', doc="Color of the wall.", native=native)

                @inject_fields(height, color)
                def __init__(self, fields):
                    with pytest.raises(MandatoryFieldInitError):
                        print(self.height)

                    # initialize all fields received
                    fields.init(self)
        else:
            class Wall:
                height: int = field(doc="Height of the wall in mm.", native=native)
                color: str = field(default='white', doc="Color of the wall.", native=native)

                @inject_fields
                def __init__(self, fields):
                    with pytest.raises(MandatoryFieldInitError):
                        print(self.height)

                    # initialize all fields received
                    fields.init(self)
    elif init_type == 'make_init':
        if explicit_fields_list:
            class Wall(object):
                height: int = field(doc="Height of the wall in mm.", native=native)
                color: str = field(default='white', doc="Color of the wall.", native=native)
                __init__ = make_init(height, color)
        else:
            class Wall(object):
                height: int = field(doc="Height of the wall in mm.", native=native)
                color: str = field(default='white', doc="Color of the wall.", native=native)
                __init__ = make_init()
    elif init_type == 'make_init_with_postinit':
        if explicit_fields_list:
            class Wall(object):
                height: int = field(doc="Height of the wall in mm.", native=native)
                color: str = field(default='white', doc="Color of the wall.", native=native)

                def post_init(self, foo: str = 'bar'):
                    print(self.height)
                    print("post init man !")
                __init__ = make_init(height, color, post_init_fun=post_init)
        else:
            class Wall(object):
                height: int = field(doc="Height of the wall in mm.", native=native)
                color: str = field(default='white', doc="Color of the wall.", native=native)

                def post_init(self, foo: str = 'bar'):
                    print(self.height)
                    print("post init man !")

                __init__ = make_init(post_init_fun=post_init)
    else:
        raise ValueError(init_type)

    return Wall


def _test_autofields(type_check):
    if type_check:
        _deco = autofields(check_types=True)
    else:
        _deco = autofields

    @_deco
    class Foo:
        CONSTANT: str = 's'
        __a__: int = 0

        foo: int
        bar = 0
        barcls = int
        barfunc = lambda x: x
        barbar: str

        class cls:
            pass

        def fct(self):
            return 1

    return Foo


def _test_autofields_readme():

    @autofields(make_init=True)
    class Item:
        name: str

    @autofields
    class Pocket:
        size: int
        items: List[Item] = []

    @autofields
    class Pocket2:
        size: int
        items: List[Item] = []
        def __init__(self, who):
            print("hello, %s" % who)

    return Pocket, Item, Pocket2


def _test_autofields_vtypes_readme():
    from vtypes import VType

    class PositiveInt(int, VType):
        __validators__ = {'should be positive': lambda x: x >= 0}

    @autofields(check_types=True)
    class Rectangle:
        x: PositiveInt
        y: PositiveInt

    return Rectangle



#  Authors: Sylvain Marie <sylvain.marie@se.com>
#
#  Copyright (c) Schneider Electric Industries, 2019. All right reserved.
import sys

import pytest


@pytest.mark.skipif(sys.version_info < (3, 6), reason="member type hints not supported in python < 3.6")
def test_issue_51():
    from ._test_py36 import test_issue_51
    A = test_issue_51()
    a = A()



#  Authors: Sylvain Marie <sylvain.marie@se.com>
#
#  Copyright (c) Schneider Electric Industries, 2020. All right reserved.
import sys

import pytest

from pyfields import get_fields


@pytest.mark.skipif(sys.version_info < (3, 6), reason="member type hints not supported in python < 3.6")
def test_issue_67():
    from ._test_py36 import test_issue_67
    Frog = test_issue_67()

    assert len(get_fields(Frog)) == 1
    Frog()



#  Authors: Sylvain Marie <sylvain.marie@se.com>
#
#  Copyright (c) Schneider Electric Industries, 2019. All right reserved.
import pytest
import dataclasses as dc
from attr import attrs, attrib

from pyfields import field, inject_fields


def _create_class_creator(type):
    if type == "pyfields":
        def _call_me():
            class Wall(object):
                height = field(doc="Height of the wall in mm.")  # type: int
                color = field(default='white', doc="Color of the wall.")  # type: str

                @inject_fields
                def __init__(self, fields):
                    fields.init(self)
            return Wall

    elif type == "python":
        def _call_me():
            class Wall(object):
                def __init__(self,
                             height,        # type: int
                             color='white'   # type: str
                             ):
                    """

                    :param height: Height of the wall in mm.
                    :param color: Color of the wall.
                    """
                    self.height = height
                    self.color = color
            return Wall
    elif type == "attrs":
        def _call_me():
            @attrs
            class Wall(object):
                height: int = attrib(repr=False, cmp=False, hash=False)
                color: str = attrib(default='white', repr=False, cmp=False, hash=False)

            return Wall

    elif type == "dataclass":
        def _call_me():
            @dc.dataclass
            class Wall(object):
                height: int = dc.field(init=True)
                color: str = dc.field(default='white', init=True)

            return Wall
    else:
        raise ValueError()

    return _call_me


@pytest.mark.parametrize("type", ["python", "pyfields", "attrs", "dataclass"])
def test_timers_class(benchmark, type):
    # benchmark it
    benchmark(_create_class_creator(type))


def _instantiate(clazz):
    return lambda: clazz(color='hello', height=50)


@pytest.mark.parametrize("type", ["python", "pyfields", "attrs", "dataclass"])
def test_timers_instance(benchmark, type):
    clazz = _create_class_creator(type)()

    benchmark(_instantiate(clazz))


def _read_field(obj):
    return lambda: obj.color


@pytest.mark.parametrize("type", ["python", "pyfields", "attrs", "dataclass"])
def test_timers_instance_read(benchmark, type):
    clazz = _create_class_creator(type)()
    obj = clazz(color='hello', height=50)

    benchmark(_read_field(obj))


def _write_field(obj):
    obj.color = 'sky_blue'


@pytest.mark.parametrize("type", ["python", "pyfields", "attrs", "dataclass"])
def test_timers_instance_write(benchmark, type):
    clazz = _create_class_creator(type)()
    obj = clazz(color='hello', height=50)

    benchmark(_write_field, obj)



#  Authors: Sylvain Marie <sylvain.marie@se.com>
#
#  Copyright (c) Schneider Electric Industries, 2019. All right reserved.

import pytest

from pyfields import ReadOnlyFieldError
from valid8 import ValidationError


def test_so0(capsys):
    """ Checks answer at https://stackoverflow.com/a/58344434/7262247 """

    from pyfields import field, init_fields

    with capsys.disabled():
        class C(object):
            x = field(default=None, doc="the optional 'x' property")
            y = field(doc="the mandatory 'y' property")
            z = field(doc="the mandatory 'z' property")

            @init_fields
            def __init__(self):
                pass

    c = C(y=1, z=2)
    print(vars(c))

    with capsys.disabled():
        out, err = capsys.readouterr()
        print(out)
        # assert out == """{'y': 1, 'x': None, 'z': 2}\n"""
        assert vars(c) == {'y': 1, 'x': None, 'z': 2}


def test_so1(capsys):
    """ Checks answer at https://stackoverflow.com/a/58344853/7262247 """

    from pyfields import field, init_fields

    class Account(object):
        first = field(doc="first name")
        last = field(doc="last name")
        age = field(doc="the age in years")
        id = field(doc="an identifier")
        balance = field(doc="current balance in euros")

        @init_fields
        def __init__(self, msg):
            print(msg)

    a = Account("hello, world!", first="s", last="marie", age=135, id=0, balance=-200000)
    print(vars(a))
    with capsys.disabled():
        out, err = capsys.readouterr()
        print(out)
        assert out.splitlines()[0] == "hello, world!"
        assert vars(a) == {'age': 135, 'balance': -200000, 'last': 'marie', 'id': 0, 'first': 's'}


def test_so2():
    """ Checks that answer at https://stackoverflow.com/a/58383062/7262247 is ok """

    from pyfields import field

    class Position(object):
        x = field(validators=lambda x: x > 0)
        y = field(validators={'y should be between 0 and 100': lambda y: y > 0 and y < 100})

    p = Position()
    p.x = 1
    with pytest.raises(ValidationError) as exc_info:
        p.y = 101
    qualname = Position.__dict__['y'].qualname
    assert str(exc_info.value) == "Error validating [%s=101]. " \
                                  "InvalidValue: y should be between 0 and 100. " \
                                  "Function [<lambda>] returned [False] for value 101." % qualname


def test_so3():
    """https://stackoverflow.com/a/58391645/7262247"""

    from pyfields import field

    class Spam(object):
        description = field(validators={"description can not be empty": lambda s: len(s) > 0})
        value = field(validators={"value must be greater than zero": lambda x: x > 0})

    s = Spam()
    with pytest.raises(ValidationError) as exc_info:
        s.description = ""
    qualname = Spam.__dict__['description'].qualname
    assert str(exc_info.value) == "Error validating [%s='']. " \
                                  "InvalidValue: description can not be empty. " \
                                  "Function [<lambda>] returned [False] for value ''." % qualname


def test_so4():
    """check https://stackoverflow.com/a/58394381/7262247"""

    from pyfields import field, init_fields
    from valid8.validation_lib import is_in

    ALLOWED_COLORS = ('blue', 'yellow', 'brown')

    class Car(object):
        """ My class with many fields """
        color = field(type_hint=str, check_type=True, validators=is_in(ALLOWED_COLORS))
        name = field(type_hint=str, check_type=True, validators={'should be non-empty': lambda s: len(s) > 0})
        wheels = field(type_hint=int, check_type=True, validators={'should be positive': lambda x: x > 0})

        @init_fields
        def __init__(self, msg="hello world!"):
            print(msg)

    c = Car(color='blue', name='roadie', wheels=3)
    assert vars(c) == {'_wheels': 3, '_name': 'roadie', '_color': 'blue'}

    qualname = Car.__dict__['wheels'].qualname

    with pytest.raises(TypeError) as exc_info:
        c.wheels = 'hello'
    assert str(exc_info.value) == "Invalid value type provided for '%s'. " \
                                  "Value should be of type %r. " \
                                  "Instead, received a 'str': 'hello'" % (qualname, int)

    with pytest.raises(ValidationError) as exc_info:
        c.wheels = 0
    assert str(exc_info.value) == "Error validating [%s=0]. " \
                                  "InvalidValue: should be positive. " \
                                  "Function [<lambda>] returned [False] for value 0." % qualname


def test_so5():
    """https://stackoverflow.com/a/58395677/7262247"""

    from pyfields import field, copy_value, init_fields
    from valid8.validation_lib import is_in

    class Well(object):
        name = field()                                        # Required
        group = field()                                       # Required
        operate_list = field(default_factory=copy_value([]))  # Optional
        monitor_list = field(default_factory=copy_value([]))  # Optional
        geometry = field(default=None)                        # Optional
        perf = field(default=None)                            # Optional

    valid_types = ('type_A', 'type_B')

    class Operate(object):
        att = field()                                  # Required
        type_ = field(type_hint=str, check_type=True, validators=is_in(valid_types))   # Required
        value = field(default_factory=copy_value([]))  # Optional
        mode = field(default=None)                     # Optional
        action = field(default=None)                   # Optional

        @init_fields
        def __init__(self):
            pass

    o = Operate(att="foo", type_='type_A')

    with pytest.raises(TypeError):
        o.type_ = 1  # <-- raises TypeError: Invalid value type provided

    with pytest.raises(ValidationError):
        bad_o = Operate(att="foo", type_='type_WRONG')  # <-- raises ValidationError: NotInAllowedValues: x in ('type_A', 'type_B') does not hold for x=type_WRONG


def test_so6():
    """checks that answer at https://stackoverflow.com/a/58396678/7262247 works"""
    from pyfields import field, init_fields

    class Position(object):
        x = field(type_hint=int, check_type=True, validators=lambda x: x > 0)
        y = field(type_hint=int, check_type=True, validators={'y should be between 0 and 100': lambda y: y > 0 and y < 100})

        @init_fields
        def __init__(self, msg="hello world!"):
            print(msg)

    p = Position(x=1, y=12)
    with pytest.raises(TypeError) as exc_info:
        p.x = '1'
    qualname = Position.__dict__['x'].qualname
    assert str(exc_info.value) == "Invalid value type provided for '%s'. " \
                                  "Value should be of type %r. Instead, received a 'str': '1'" % (qualname, int)

    with pytest.raises(ValidationError) as exc_info:
        p.y = 101


def test_so7():
    """ checks answer at https://stackoverflow.com/a/58432813/7262247 """

    from pyfields import field

    class User(object):
        username = field(read_only=True, validators={'should contain more than 2 characters': lambda s: len(s) > 2})

    u = User()
    u.username = "earthling"
    assert vars(u) == {'_username': "earthling"}
    with pytest.raises(ReadOnlyFieldError) as exc_info:
        u.username = "earthling2"
    qualname = User.__dict__['username'].qualname
    assert str(exc_info.value) == "Read-only field '%s' has already been initialized on instance %s and cannot be " \
                                  "modified anymore." % (qualname, u)



#  Authors: Sylvain Marie <sylvain.marie@se.com>
#
#  Copyright (c) Schneider Electric Industries, 2019. All right reserved.
from copy import copy, deepcopy
from inspect import getmro

try:
    from typing import Union, Type, TypeVar
    T = TypeVar('T')
except ImportError:
    pass

from pyfields.core import Field, ClassFieldAccessError, PY36, get_type_hints


class NotAFieldError(TypeError):
    """ Raised by `get_field` when the class member with that name is not a field """
    __slots__ = 'name', 'cls'

    def __init__(self, cls, name):
        self.name = name
        self.cls = cls


def get_field(cls, name):
    """
    Utility method to return the field member with name `name` in class `cls`.
    If the member is not a field, a `NotAFieldError` is raised.

    :param cls:
    :param name:
    :return:
    """
    try:
        member = getattr(cls, name)
    except ClassFieldAccessError as e:
        # we know it is a field :)
        return e.field
    except Exception:
        # any other exception that can happen with a descriptor for example
        raise NotAFieldError(cls, name)
    else:
        # it is a field if is it an instance of Field
        if isinstance(member, Field):
            return member
        else:
            raise NotAFieldError(cls, name)


def yield_fields(cls,
                 include_inherited=True,  # type: bool
                 remove_duplicates=True,  # type: bool
                 ancestors_first=True,    # type: bool
                 _auto_fix_fields=False   # type: bool
                 ):
    """
    Similar to `get_fields` but as a generator.

    :param cls:
    :param include_inherited:
    :param remove_duplicates:
    :param ancestors_first:
    :param _auto_fix_fields:
    :return:
    """
    # List the classes where we should be looking for fields
    if include_inherited:
        where_cls = reversed(getmro(cls)) if ancestors_first else getmro(cls)
    else:
        where_cls = (cls,)

    # Init
    _already_found_names = set() if remove_duplicates else None  # a reference set of already yielded field names
    _cls_pep484_member_type_hints = None                         # where to hold found type hints if needed
    _all_fields_for_cls = None                                   # temporary list when we have to reorder

    # finally for each class, gather all fields in order
    for _cls in where_cls:
        if not PY36:
            # in python < 3.6 we'll need to sort the fields at the end as class member order is not preserved
            _all_fields_for_cls = []
        elif _auto_fix_fields:
            # in python >= 3.6, pep484 type hints can be available as member annotation, grab them
            _cls_pep484_member_type_hints = get_type_hints(_cls)

        for member_name in vars(_cls):
            # if not member_name.startswith('__'):   not stated in the doc: too dangerous to have such implicit filter

            # avoid infinite recursion as this method is called in the descriptor for __init__
            if not member_name == '__init__':
                try:
                    field = get_field(_cls, member_name)
                except NotAFieldError:
                    continue

                if _auto_fix_fields:
                    # take this opportunity to set the name and type hints
                    field.set_as_cls_member(_cls, member_name, owner_cls_type_hints=_cls_pep484_member_type_hints)

                if remove_duplicates:
                    if member_name in _already_found_names:
                        continue
                    else:
                        _already_found_names.add(member_name)

                # maybe the field is overridden, in that case we should directly yield the new one
                if _cls is not cls:
                    try:
                        overridden_field = get_field(cls, member_name)
                    except NotAFieldError:
                        overridden_field = None
                else:
                    overridden_field = None

                # finally yield it...
                if PY36:  # ...immediately in recent python versions because order is correct already
                    yield field if overridden_field is None else overridden_field
                else:     # ...or wait for this class to be collected, because the order needs to be fixed
                    _all_fields_for_cls.append((field, overridden_field))

        if not PY36:
            # order is random in python < 3.6 - we need to explicitly sort according to instance creation number
            _all_fields_for_cls.sort(key=lambda f: f[0].__fieldinstcount__)
            for field, overridden_field in _all_fields_for_cls:
                yield field if overridden_field is None else overridden_field


def has_fields(cls,
               include_inherited=True  # type: bool
               ):
    """
    Returns True if class `cls` defines at least one `pyfields` field.
    If `include_inherited` is `True` (default), the method will return `True` if at least a field is defined in the
    class or one of its ancestors. If `False`, the fields need to be defined on the class itself.

    :param cls:
    :param include_inherited:
    :return:
    """
    return any(yield_fields(cls, include_inherited=include_inherited))


def get_fields(cls,
               include_inherited=True,  # type: bool
               remove_duplicates=True,  # type: bool
               ancestors_first=True,    # type: bool
               container_type=tuple,    # type: Type[T]
               _auto_fix_fields=False   # type: bool
               ):
    # type: (...) -> T
    """
    Utility method to collect all fields defined in a class, including all inherited or not.

    By default duplicates are removed and ancestor fields are included and appear first. If a field is overridden,
    it will appear at the position of the overridden field in the order.

    :param cls:
    :param include_inherited:
    :param remove_duplicates:
    :param ancestors_first:
    :param container_type:
    :param _auto_fix_fields:
    :return: the fields (by default, as a tuple)
    """
    return container_type(yield_fields(cls, include_inherited=include_inherited,
                                       remove_duplicates=remove_duplicates, ancestors_first=ancestors_first,
                                       _auto_fix_fields=_auto_fix_fields))


# def ordered_dir(cls,
#                 ancestors_first=False  # type: bool
#                 ):
#     """
#     since `dir` does not preserve order, lets have our own implementation
#
#     :param cls:
#     :param ancestors_first:
#     :return:
#     """
#     classes = reversed(getmro(cls)) if ancestors_first else getmro(cls)
#
#     for _cls in classes:
#         for k in vars(_cls):
#             yield k


def copy_value(val,
               deep=True  # type: bool
               ):
    """
    Returns a default value factory to be used in a `field(default_factory=...)`.

    That factory will create a copy of the provided `val` everytime it is called. Handy if you wish to use mutable
    objects as default values for your fields ; for example lists.

    :param val: the (mutable) value to copy
    :param deep: by default deep copies will be created. You can change this behaviour by setting this to `False`
    :return:
    """
    if deep:
        def create_default(obj):
            return deepcopy(val)
    else:
        def create_default(obj):
            return copy(val)

    # attach a method to easily get a new factory with a new value
    def get_copied_value():
        return val

    def clone_with_new_val(newval):
        return copy_value(newval, deep)

    create_default.get_copied_value = get_copied_value
    create_default.clone_with_new_val = clone_with_new_val
    return create_default


def copy_field(field_or_name,  # type: Union[str, Field]
               deep=True       # type: bool
               ):
    """
    Returns a default value factory to be used in a `field(default_factory=...)`.

    That factory will create a copy of the value in the given field. You can provide a field or a field name, in which
    case this method is strictly equivalent to `copy_attr`.

    :param field_or_name: the field or name of the field for which the value needs to be copied
    :param deep: by default deep copies will be created. You can change this behaviour by setting this to `False`
    :return:
    """
    if isinstance(field_or_name, Field):
        if field_or_name.name is None:
            # Name not yet available, we'll get it later
            if deep:
                def create_default(obj):
                    return deepcopy(getattr(obj, field_or_name.name))
            else:
                def create_default(obj):
                    return copy(getattr(obj, field_or_name.name))

            return create_default
        else:
            # use the field name
            return copy_attr(field_or_name.name, deep=deep)
    else:
        # this is already a field name
        return copy_attr(field_or_name, deep=deep)


def copy_attr(attr_name,  # type: str
              deep=True   # type: bool
              ):
    """
    Returns a default value factory to be used in a `field(default_factory=...)`.

    That factory will create a copy of the value in the given attribute.

    :param attr_name: the name of the attribute for which the value will be copied
    :param deep: by default deep copies will be created. You can change this behaviour by setting this to `False`
    :return:
    """
    if deep:
        def create_default(obj):
            return deepcopy(getattr(obj, attr_name))
    else:
        def create_default(obj):
            return copy(getattr(obj, attr_name))

    return create_default



#  Authors: Sylvain Marie <sylvain.marie@se.com>
#
#  Copyright (c) Schneider Electric Industries, 2019. All right reserved.
import sys
from inspect import isfunction, getmro
from itertools import islice

try:
    from inspect import signature, Parameter, Signature
except ImportError:
    from funcsigs import signature, Parameter, Signature


try:  # python 3.5+
    from typing import Optional, Set, List, Callable, Dict, Type, Any, TypeVar, Union, Iterable, Tuple, Mapping
    from valid8.common_syntax import ValidationFuncs
    use_type_hints = sys.version_info > (3, 0)
except ImportError:
    use_type_hints = False


from makefun import wraps, with_signature

from pyfields.core import PY36, USE_FACTORY, EMPTY, Field
from pyfields.helpers import get_fields


def init_fields(*fields,   # type: Union[Field, Any]
                **kwargs
                ):
    """
    Decorator for an init method, so that fields are initialized before entering the method.

    By default, when the decorator is used without arguments or when `fields` is empty, all fields defined in the class
    are initialized. Fields inherited from parent classes are included, following the mro. The signature of the init
    method is modified so that it can receive values for these fields:

    >>> import sys, pytest
    >>> if sys.version_info < (3, 7): pytest.skip('doctest skipped')  # 3.6 help() is different on travis

    >>> from pyfields import field, init_fields
    >>> class Wall:
    ...     height: int = field(doc="Height of the wall in mm.")
    ...     color: str = field(default='white', doc="Color of the wall.")
    ...
    ...     @init_fields
    ...     def __init__(self, msg: str = 'hello'):
    ...         print("post init ! height=%s, color=%s, msg=%s" % (self.height, self.color, msg))
    ...         self.non_field_attr = msg
    ...
    >>> help(Wall.__init__)
    Help on function __init__ in module pyfields.init_makers:
    <BLANKLINE>
    __init__(self, height: int, msg: str = 'hello', color: str = 'white')
        The `__init__` method generated for you when you use `@init_fields`
        or `make_init` with a non-None `post_init_fun` method.
    <BLANKLINE>
    >>> w = Wall(2)
    post init ! height=2, color=white, msg=hello


    The list of fields can be explicitly provided in `fields`.

    By default the init arguments will appear before the fields in the signature, wherever possible (mandatory args
    before mandatory fields, optional args before optional fields). You can change this behaviour by setting
    `init_args_before` to `False`.

    :param fields: list of fields to initialize before entering the decorated `__init__` method. For each of these
        fields a corresponding argument will be added in the method's signature. If an empty list is provided, all
        fields from the class will be used including inherited fields following the mro. In case of inherited fields,
        they will appear before the class fields if `ancestor_fields_first` is `True` (default), after otherwise.
    :param init_args_before: If set to `True` (default), arguments from the decorated init method will appear before
        the fields when possible. If set to `False` the contrary will happen.
    :param ancestor_fields_first: If set to `True` (default behaviour), when the provided list of fields is empty,
        ancestor-inherited fields will appear before the class fields when possible (even for fields overridden in the
        subclass). If set to `False` the contrary will happen.
    :return:
    """
    init_args_before, ancestor_fields_first = pop_kwargs(kwargs, [('init_args_before', True),
                                                                  ('ancestor_fields_first', None)], allow_others=False)

    if len(fields) == 1:
        # used without argument ?
        f = fields[0]
        if isfunction(f) and not isinstance(f, Field) and init_args_before:
            # @init_fields decorator used without parenthesis

            # The list of fields is NOT explicit: we have no way to gather this list without creating a descriptor
            return InitDescriptor(user_init_fun=f, user_init_is_injected=False,
                                  ancestor_fields_first=ancestor_fields_first)

    def apply_decorator(init_fun):
        # @init_fields(...)

        # The list of fields is explicit AND names/type hints have been set already:
        # it is not easy to be sure of this (names yes, but annotations?) > prefer the descriptor anyway
        return InitDescriptor(fields=fields, user_init_fun=init_fun, user_init_args_before=init_args_before,
                              user_init_is_injected=False, ancestor_fields_first=ancestor_fields_first)

    return apply_decorator


def inject_fields(*fields  # type: Union[Field, Any]
                  ):
    """
    A decorator for `__init__` methods, to make them automatically expose arguments corresponding to all `*fields`.
    It can be used with or without arguments. If the list of fields is empty, it means "all fields from the class".

    The decorated `__init__` method should have an argument named `'fields'`. This argument will be injected with an
    object so that users can manually execute the fields initialization. This is done with `fields.init()`.

    >>> import sys, pytest
    >>> if sys.version_info < (3, 6): pytest.skip('doctest skipped')

    >>> from pyfields import field, inject_fields
    ...
    >>> class Wall(object):
    ...     height = field(doc="Height of the wall in mm.")
    ...     color = field(default='white', doc="Color of the wall.")
    ...
    ...     @inject_fields(height, color)
    ...     def __init__(self, fields):
    ...         # initialize all fields received
    ...         fields.init(self)
    ...
    ...     def __repr__(self):
    ...         return "Wall<height=%r, color=%r>" % (self.height, self.color)
    ...
    >>> Wall()
    Traceback (most recent call last):
    ...
    TypeError: __init__() missing 1 required positional argument: 'height'
    >>> Wall(1)
    Wall<height=1, color='white'>

    :param fields: list of fields to initialize before entering the decorated `__init__` method. For each of these
        fields a corresponding argument will be added in the method's signature. If an empty list is provided, all
        fields from the class will be used including inherited fields following the mro.
    :return:
    """
    if len(fields) == 1:
        # used without argument ?
        f = fields[0]
        if isfunction(f) and not isinstance(f, Field):
            # @inject_fields decorator used without parenthesis

            # The list of fields is NOT explicit: we have no way to gather this list without creating a descriptor
            return InitDescriptor(user_init_fun=f, user_init_is_injected=True)

    def apply_decorator(init_fun):
        # @inject_fields(...)

        # The list of fields is explicit AND names/type hints have been set already:
        # it is not easy to be sure of this (names yes, but annotations?) > prefer the descriptor anyway
        return InitDescriptor(user_init_fun=init_fun, fields=fields, user_init_is_injected=True)

    return apply_decorator


def make_init(*fields,  # type: Union[Field, Any]
              **kwargs
              ):
    # type: (...) -> InitDescriptor
    """
    Creates a constructor based on the provided fields.

    If `fields` is empty, all fields from the class will be used in order of appearance, then the ancestors (following
    the mro)

    >>> import sys, pytest
    >>> if sys.version_info < (3, 6): pytest.skip('doctest skipped')

    >>> from pyfields import field, make_init
    ...
    >>> class Wall:
    ...     height = field(doc="Height of the wall in mm.")
    ...     color = field(default='white', doc="Color of the wall.")
    ...     __init__ = make_init()
    >>> w = Wall(1, color='blue')
    >>> assert vars(w) == {'color': 'blue', 'height': 1}

   If `fields` is not empty, only the listed fields will appear in the constructor and will be initialized upon init.

    >>> class Wall:
    ...     height = field(doc="Height of the wall in mm.")
    ...     color = field(default='white', doc="Color of the wall.")
    ...     __init__ = make_init(height)
    >>> w = Wall(1, color='blue')
    Traceback (most recent call last):
    ...
    TypeError: __init__() got an unexpected keyword argument 'color'

    `fields` can contain fields that do not belong to this class: typically they can be fields defined in a parent
    class. Note however that any field can be used, it is not mandatory to use class or inherited fields.

    >>> class Wall:
    ...     height: int = field(doc="Height of the wall in mm.")
    ...
    >>> class ColoredWall(Wall):
    ...     color: str = field(default='white', doc="Color of the wall.")
    ...     __init__ = make_init(Wall.__dict__['height'])
    ...
    >>> w = ColoredWall(1)
    >>> vars(w)
    {'height': 1}

    If a `post_init_fun` is provided, it should be a function with `self` as first argument. This function will be
    executed after all declared fields have been initialized. The signature of the resulting `__init__` function
    created will be constructed by blending all mandatory/optional fields with the mandatory/optional args in the
    `post_init_fun` signature. The ones from the `post_init_fun` will appear first except if `post_init_args_before`
    is set to `False`

    >>> class Wall:
    ...     height: int = field(doc="Height of the wall in mm.")
    ...     color: str = field(default='white', doc="Color of the wall.")
    ...
    ...     def post_init(self, msg='hello'):
    ...         print("post init ! height=%s, color=%s, msg=%s" % (self.height, self.color, msg))
    ...         self.non_field_attr = msg
    ...
    ...     # only `height` and `foo` will be in the constructor
    ...     __init__ = make_init(height, post_init_fun=post_init)
    ...
    >>> w = Wall(1, 'hey')
    post init ! height=1, color=white, msg=hey
    >>> assert vars(w) == {'height': 1, 'color': 'white', 'non_field_attr': 'hey'}

    :param fields: the fields to include in the generated constructor signature. If no field is provided, all fields
        defined in the class will be included, as well as inherited ones following the mro.
    :param post_init_fun: (default: `None`) an optional function to call once all fields have been initialized. This
        function should have `self` as first argument. The rest of its signature will be blended with the fields in the
        generated constructor signature.
    :param post_init_args_before: boolean. Defines if the arguments from the `post_init_fun` should appear before
        (default: `True`) or after (`False`) the fields in the generated signature. Of course in all cases, mandatory
        arguments will appear after optional arguments, so as to ensure that the created signature is valid.
    :param ancestor_fields_first: If set to `True` (default behaviour), when the provided list of fields is empty,
        ancestor-inherited fields will appear before the class fields when possible (even for fields overridden in the
        subclass). If set to `False` the contrary will happen.
    :return: a constructor method to be used as `__init__`
    """
    # python <3.5 compliance: pop the kwargs following the varargs
    post_init_fun, post_init_args_before, ancestor_fields_first = pop_kwargs(kwargs, [
        ('post_init_fun', None), ('post_init_args_before', True), ('ancestor_fields_first', None)
    ], allow_others=False)

    return InitDescriptor(fields=fields, user_init_fun=post_init_fun, user_init_args_before=post_init_args_before,
                          user_init_is_injected=False, ancestor_fields_first=ancestor_fields_first)


class InitDescriptor(object):
    """
    A class member descriptor for the `__init__` method that we create with `make_init`.
    The first time people access `cls.__init__`, the actual method will be created and injected in the class.
    This descriptor will then disappear and the class will behave normally.

    The reason why we do not create the init method directly is that we require all fields to be attached to the class
    so that they have names and type hints.

    Inspired by https://stackoverflow.com/a/3412743/7262247
    """
    __slots__ = 'fields', 'user_init_is_injected', 'user_init_fun', 'user_init_args_before', 'ancestor_fields_first', \
                'ownercls'

    def __init__(self, fields=None, user_init_is_injected=False, user_init_fun=None, user_init_args_before=True,
                 ancestor_fields_first=None):
        if fields is not None and len(fields) == 0:
            fields = None
        self.fields = fields
        self.user_init_is_injected = user_init_is_injected
        self.user_init_fun = user_init_fun
        self.user_init_args_before = user_init_args_before
        if ancestor_fields_first is None:
            ancestor_fields_first = True
        elif fields is not None:
            raise ValueError("`ancestor_fields_first` is only applicable when `fields` is empty")
        self.ancestor_fields_first = ancestor_fields_first
        self.ownercls = None

    def __set_name__(self, owner, name):
        """
        There is a python issue with init descriptors with super() access. To fix it we need to
        remember the owner class type separately as we cant' trust the one received in __get__.
        See https://github.com/smarie/python-pyfields/issues/53
        """
        self.ownercls = owner

    def __get__(self, obj, objtype):
        # type: (...) -> Callable
        """
        THIS IS NOT THE INIT METHOD ! THIS IS THE CREATOR OF THE INIT METHOD (first time only)
        Python Descriptor protocol - this is called when the __init__ method is required for the first time,
        it creates the `__init__` method, replaces itself with it, and returns it. Subsequent calls will directly
        be routed to the new init method and not here.
        """
        # objtype is not reliable: when called through super() it does not contain the right class.
        # see https://github.com/smarie/python-pyfields/issues/53
        if self.ownercls is not None:
            objtype = self.ownercls
        elif objtype is not None:
            # workaround in case of python < 3.6: at least, when a subclass init is created, make sure that all super
            # classes init have their owner class properly set, .
            # That way, when the subclass __init__ will be called, containing potential calls to super(), the parents'
            # __init__ method descriptors will be correctly configured.
            for _c in reversed(getmro(objtype)[1:-1]):
                try:
                    _init_member = _c.__dict__['__init__']
                except KeyError:
                    continue
                else:
                    if isinstance(_init_member, InitDescriptor):
                        if _init_member.ownercls is None:
                            # call __set_name__ explicitly (python < 3.6) to register the descriptor with the class
                            _init_member.__set_name__(_c, '__init__')

        # <objtype>.__init__ has been accessed. Create the modified init
        fields = self.fields
        if fields is None:
            # fields have not been provided explicitly, collect them all.
            fields = get_fields(objtype, include_inherited=True, ancestors_first=self.ancestor_fields_first,
                                _auto_fix_fields=not PY36)
        elif not PY36:
            # take this opportunity to apply all field names including inherited
            # TODO set back inherited = False when the bug with class-level access is solved -> make_init will be ok
            get_fields(objtype, include_inherited=True, ancestors_first=self.ancestor_fields_first,
                       _auto_fix_fields=True)

        # create the init method
        new_init = create_init(fields=fields, inject_fields=self.user_init_is_injected,
                               user_init_fun=self.user_init_fun, user_init_args_before=self.user_init_args_before)

        # replace it forever in the class
        setattr(objtype, '__init__', new_init)

        # return the new init
        return new_init.__get__(obj, objtype)


class InjectedInitFieldsArg(object):
    """
    The object that is injected in the users' `__init__` method as the `fields` argument,
    when it has been decorated with `@inject_fields`.

    All field values received from the generated `__init__` are available in `self.field_values`, and
    a `init()` method allows users to perform the initialization per se.
    """
    __slots__ = 'field_values'

    def __init__(self, **init_field_values):
        self.field_values = init_field_values

    def init(self, obj):
        """
        Initializes all fields on the provided object
        :param obj:
        :return:
        """
        for field_name, field_value in self.field_values.items():
            if field_value is not USE_FACTORY:
                # init the field with the provided value or the injected default value
                setattr(obj, field_name, field_value)
            else:
                # init the field with its factory
                getattr(obj, field_name)


def create_init(fields,                     # type: Iterable[Field]
                user_init_fun=None,         # type: Callable[[...], Any]
                inject_fields=False,        # type: bool
                user_init_args_before=True  # type: bool
                ):
    """
    Creates the new init function that will replace `init_fun`.
    It requires that all fields have correct names and type hints so we usually execute it from within a __init__
    descriptor.

    :param fields:
    :param user_init_fun:
    :param inject_fields:
    :param user_init_args_before:
    :return:
    """
    # the list of parameters that should be exposed
    params = [Parameter('self', kind=Parameter.POSITIONAL_OR_KEYWORD)]

    if user_init_fun is None:
        # A - no function provided: expose a signature containing 'self' + fields
        field_names, _ = _insert_fields_at_position(fields, params, 1)
        new_sig = Signature(parameters=params)

        # and create the new init method
        @with_signature(new_sig, func_name='__init__')
        def init_fun(*args, **kwargs):
            """
            The `__init__` method generated for you when you use `make_init`
            """
            # 1. get 'self'
            try:
                # most of the time 'self' will be received like that
                self = kwargs['self']
            except IndexError:
                self = args[0]

            # 2. self-assign all fields
            for field_name in field_names:
                field_value = kwargs[field_name]
                if field_value is not USE_FACTORY:
                    # init the field with the provided value or the injected default value
                    setattr(self, field_name, field_value)
                else:
                    # init the field with its factory, by just getting it
                    getattr(self, field_name)

        return init_fun

    else:
        # B - function provided - expose a signature containing 'self' + the function params + fields
        # start by inserting all fields
        field_names, _idx = _insert_fields_at_position(fields, params, 1)

        # then get the function signature
        user_init_sig = signature(user_init_fun)

        # Insert all parameters from the function except 'self'
        if user_init_args_before:
            mandatory_insert_idx, optional_insert_idx = 1, _idx
        else:
            mandatory_insert_idx, optional_insert_idx = _idx, len(params)

        fields_arg_found = False
        for p in islice(user_init_sig.parameters.values(), 1, None):  # remove the 'self' argument
            if inject_fields and p.name == 'fields':
                # injected argument
                fields_arg_found = True
                continue
            if p.default is p.empty:
                # mandatory
                params.insert(mandatory_insert_idx, p)
                mandatory_insert_idx += 1
                optional_insert_idx += 1
            else:
                # optional
                params.insert(optional_insert_idx, p)
                optional_insert_idx += 1

        if inject_fields and not fields_arg_found:
            # 'fields' argument not found in __init__ signature: impossible to inject, raise an error
            try:
                name = user_init_fun.__qualname__
            except AttributeError:
                name = user_init_fun.__name__
            raise ValueError("Error applying `@inject_fields` on `%s%s`: "
                             "no 'fields' argument is available in the signature." % (name, user_init_sig))

        # replace the signature with the newly created one
        new_sig = user_init_sig.replace(parameters=params)

        # and create the new init method
        if inject_fields:
            @wraps(user_init_fun, new_sig=new_sig)
            def __init__(self, *args, **kwargs):
                """
                The `__init__` method generated for you when you use `@inject_fields` on your `__init__`
                """
                # 1. remove all field values received from the outer signature
                _fields = dict()
                for f_name in field_names:
                    _fields[f_name] = kwargs.pop(f_name)

                # 2. inject our special variable
                kwargs['fields'] = InjectedInitFieldsArg(**_fields)

                # 3. call your __init__ method
                return user_init_fun(self, *args, **kwargs)

        else:
            @wraps(user_init_fun, new_sig=new_sig)
            def __init__(self, *args, **kwargs):
                """
                The `__init__` method generated for you when you use `@init_fields`
                or `make_init` with a non-None `post_init_fun` method.
                """
                # 1. self-assign all fields
                for field_name in field_names:
                    field_value = kwargs.pop(field_name)
                    if field_value is not USE_FACTORY:
                        # init the field with the provided value or the injected default value
                        setattr(self, field_name, field_value)
                    else:
                        # init the field with its factory, by just getting it
                        getattr(self, field_name)

                # 2. call your post-init method
                return user_init_fun(self, *args, **kwargs)

        return __init__


def _insert_fields_at_position(fields_to_insert,
                               params,
                               i,
                               field_names=None
                               ):
    """
    Note: preserve order as much as possible, but automatically place all mandatory fields first so that the
    signature is valid.

    :param fields_to_insert:
    :param field_names:
    :param i:
    :param params:
    :return:
    """
    if field_names is None:
        field_names = []

    initial_i = i
    last_mandatory_idx = i
    for _field in reversed(fields_to_insert):
        # Is this field optional ?
        if _field.is_mandatory:
            # mandatory
            where_to_insert = i
            last_mandatory_idx += 1
            default = Parameter.empty
        elif _field.is_default_factory:
            # optional with a default value factory: place a specific symbol in the signature to indicate it
            default = USE_FACTORY
            where_to_insert = last_mandatory_idx
        else:
            # optional with a default value
            default = _field.default
            where_to_insert = last_mandatory_idx

        # Are there annotations on the field ?
        annotation = _field.type_hint if _field.type_hint is not EMPTY else Parameter.empty

        # remember the list of field names for later use - but in the right order
        field_names.insert(where_to_insert - initial_i, _field.name)

        # finally inject the new parameter in the signature
        new_param = Parameter(_field.name, kind=Parameter.POSITIONAL_OR_KEYWORD, default=default, annotation=annotation)
        params.insert(where_to_insert, new_param)

    return field_names, last_mandatory_idx


def pop_kwargs(kwargs,
               names_with_defaults,  # type: List[Tuple[str, Any]]
               allow_others=False
               ):
    """
    Internal utility method to extract optional arguments from kwargs.

    :param kwargs:
    :param names_with_defaults:
    :param allow_others: if False (default) then an error will be raised if kwargs still contains something at the end.
    :return:
    """
    all_arguments = []
    for name, default_ in names_with_defaults:
        try:
            val = kwargs.pop(name)
        except KeyError:
            val = default_
        all_arguments.append(val)

    if not allow_others and len(kwargs) > 0:
        raise ValueError("Unsupported arguments: %s" % kwargs)

    if len(names_with_defaults) == 1:
        return all_arguments[0]
    else:
        return all_arguments



#  Authors: Sylvain Marie <sylvain.marie@se.com>
#
#  Copyright (c) Schneider Electric Industries, 2019. All right reserved.
import sys
from collections import OrderedDict

from valid8 import Validator, failure_raiser, ValidationError, ValidationFailure
from valid8.base import getfullargspec as v8_getfullargspec, get_callable_name, is_mini_lambda
from valid8.common_syntax import FunctionDefinitionError, make_validation_func_callables
from valid8.composition import _and_
from valid8.entry_points import _add_none_handler
from valid8.utils.signature_tools import IsBuiltInError
from valid8.validation_lib import instance_of

try:  # python 3.5+
    # noinspection PyUnresolvedReferences
    from typing import Callable, Type, Any, TypeVar, Union, Iterable, Tuple, Mapping, Optional, Dict, Literal
    from valid8.common_syntax import ValidationFuncs
    use_type_hints = sys.version_info > (3, 0)
except ImportError:
    use_type_hints = False


if use_type_hints:
    T = TypeVar('T')

    # ------------- validator type hints -----------
    # 1. the lowest-level user or 3d party-provided validation functions
    ValidationFunc = Union[Callable[[Any], Any],
                           Callable[[Any, Any], Any],
                           Callable[[Any, Any, Any], Any]]
    """A validation function is a callable with signature (val), (obj, val) or (obj, field, val), returning `True` 
    or `None` in case of success"""

    try:
        # noinspection PyUnresolvedReferences
        from mini_lambda import y
        ValidationFuncOrLambda = Union[ValidationFunc, type(y)]
    except ImportError:
        ValidationFuncOrLambda = ValidationFunc

    # 2. the syntax to optionally transform them into failure raisers by providing a tuple
    ValidatorDef = Union[ValidationFuncOrLambda,
                         Tuple[ValidationFuncOrLambda, str],
                         Tuple[ValidationFuncOrLambda, Type[ValidationFailure]],
                         Tuple[ValidationFuncOrLambda, str, Type[ValidationFailure]]
                         ]
    """A validator is a validation function together with optional error message and error type"""

    # 3. the syntax to describe several validation functions at once
    VFDefinitionElement = Union[str, Type[ValidationFailure], ValidationFuncOrLambda]
    """This type represents one of the elements that can define a checker: help msg, failure type, callable"""

    OneOrSeveralVFDefinitions = Union[ValidatorDef,
                                      Iterable[ValidatorDef],
                                      Mapping[VFDefinitionElement, Union[VFDefinitionElement,
                                                                         Tuple[VFDefinitionElement, ...]]]]
    """Several validators can be provided as a singleton, iterable, or dict-like. In that case the value can be a 
    single variable or a tuple, and it will be combined with the key to form the validator. So you can use any of 
    the elements defining a validators as the key."""

    # shortcut name used everywhere. Less explicit
    Validators = OneOrSeveralVFDefinitions


class FieldValidator(Validator):
    """
    Represents a `Validator` responsible to validate a `field`
    """
    __slots__ = '__weakref__', 'validated_field', 'base_validation_funcs'

    def __init__(self,
                 validated_field,   # type: 'DescriptorField'
                 validators,        # type: Validators
                 **kwargs
                 ):
        """

        :param validated_field: the field being validated.
        :param validators: the base validation function or list of base validation functions to use. A callable, a
            tuple(callable, help_msg_str), a tuple(callable, failure_type), tuple(callable, help_msg_str, failure_type)
            or a list of several such elements. A dict can also be used.
            Tuples indicate an implicit `failure_raiser`.
            [mini_lambda](https://smarie.github.io/python-mini-lambda/) expressions can be used instead
            of callables, they will be transformed to functions automatically.
        :param error_type: a subclass of ValidationError to raise in case of validation failure. By default a
            ValidationError will be raised with the provided help_msg
        :param help_msg: an optional help message to be used in the raised error in case of validation failure.
        :param none_policy: describes how None values should be handled. See `NonePolicy` for the various possibilities.
            Default is `NonePolicy.VALIDATE`, meaning that None values will be treated exactly like other values and
            follow the same validation process.
        :param kw_context_args: optional contextual information to store in the exception, and that may be also used
            to format the help message
        """
        # store this additional info about the function been validated
        self.validated_field = validated_field

        try:  # dict ?
            validators.keys()
        except (AttributeError, FunctionDefinitionError):  # FunctionDefinitionError when mini_lambda
            if isinstance(validators, tuple):
                # single tuple
                validators = (validators,)
            else:
                try:  # iterable
                    iter(validators)
                except (TypeError, FunctionDefinitionError):  # FunctionDefinitionError when mini_lambda
                    # single
                    validators = (validators,)
        else:
            # dict
            validators = (validators,)

        # remember validation funcs so that we can add more later
        self.base_validation_funcs = validators

        super(FieldValidator, self).__init__(*validators, **kwargs)

    def add_validator(self,
                      validation_func  # type: ValidatorDef
                      ):
        """
        Adds the provided validation function to the existing list of validation functions
        :param validation_func:
        :return:
        """
        self.base_validation_funcs = self.base_validation_funcs + (validation_func, )

        # do the same than in super.init, once again. TODO optimize ...
        validation_funcs = make_validation_func_callables(*self.base_validation_funcs,
                                                          callable_creator=self.get_callables_creator())
        main_val_func = _and_(validation_funcs)
        self.main_function = _add_none_handler(main_val_func, none_policy=self.none_policy)

    def get_callables_creator(self):
        def make_validator_callable(validation_callable,  # type: ValidationFunc
                                    help_msg=None,        # type: str
                                    failure_type=None,    # type: Type[ValidationFailure]
                                    **kw_context_args):
            """

            :param validation_callable:
            :param help_msg: custom help message for failures to raise
            :param failure_type: type of failures to raise
            :param kw_context_args: contextual arguments for failures to raise
            :return:
            """
            if is_mini_lambda(validation_callable):
                validation_callable = validation_callable.as_function()

            # support several cases for the validation function signature
            # `f(val)`, `f(obj, val)` or `f(obj, field, val)`
            # the validation function has two or three (or more but optional) arguments.
            # valid8 requires only 1.
            try:
                args, varargs, varkwargs, defaults = v8_getfullargspec(validation_callable, skip_bound_arg=True)[0:4]

                nb_args = len(args) if args is not None else 0
                nbvarargs = 1 if varargs is not None else 0
                # nbkwargs = 1 if varkwargs is not None else 0
                # nbdefaults = len(defaults) if defaults is not None else 0
            except IsBuiltInError:
                # built-ins: TypeError: <built-in function isinstance> is not a Python function
                # assume signature with a single positional argument
                nb_args = 1
                nbvarargs = 0
                # nbkwargs = 0
                # nbdefaults = 0

            if nb_args == 0 and nbvarargs == 0:
                raise ValueError(
                    "validation function should accept 1, 2, or 3 arguments at least. `f(val)`, `f(obj, val)` or "
                    "`f(obj, field, val)`")
            elif nb_args == 1 or (nb_args == 0 and nbvarargs >= 1):  # varargs default to one argument (compliance with old mini lambda)
                # `f(val)`
                def new_validation_callable(val, **ctx):
                    return validation_callable(val)
            elif nb_args == 2:
                # `f(obj, val)`
                def new_validation_callable(val, **ctx):
                    return validation_callable(ctx['obj'], val)
            else:
                # `f(obj, field, val, *opt_args, **ctx)`
                def new_validation_callable(val, **ctx):
                    # note: field is available both from **ctx and self. Use the "fastest" way
                    return validation_callable(ctx['obj'], self.validated_field, val)

            # preserve the name
            new_validation_callable.__name__ = get_callable_name(validation_callable)

            return failure_raiser(new_validation_callable, help_msg=help_msg, failure_type=failure_type,
                                  **kw_context_args)

        return make_validator_callable

    def get_additional_info_for_repr(self):
        # type: (...) -> str
        return 'validated_field=%s' % self.validated_field.qualname

    def _get_name_for_errors(self,
                             name  # type: str
                             ):
        # type: (...) -> str
        """ override this so that qualname is only called if an error is raised, not before """
        return self.validated_field.qualname

    def assert_valid(self,
                     obj,              # type: Any
                     value,            # type: Any
                     error_type=None,  # type: Type[ValidationError]
                     help_msg=None,    # type: str
                     **ctx):
        # do not use qualname here so as to save time.
        super(FieldValidator, self).assert_valid(self.validated_field.name, value,
                                                 error_type=error_type, help_msg=help_msg,
                                                 # context info contains obj and field
                                                 obj=obj, field=self.validated_field, **ctx)


# --------------- converters
supported_syntax = 'a Converter, a conversion callable, a tuple(validation_callable, conversion_callable), ' \
                   'a tuple(valid_type, conversion_callable), or a list of several such elements. ' \
                   'A special string \'*\' can be used to denote that all values are accepted.' \
                   'Dicts are supported too, in which case the key is the validation callable or the valid type.' \
                   '[mini_lambda](https://smarie.github.io/python-mini-lambda/) expressions can be used instead of ' \
                   'callables, they will be transformed to functions automatically.'


class Converter(object):
    """
    A converter to be used in `field`s.
    """
    __slots__ = ('name', )

    def __init__(self, name=None):
        self.name = name

    def __str__(self):
        if self.name is not None:
            return self.name
        else:
            return self.__class__.__name__

    def accepts(self, obj, field, value):
        # type: (...) -> Optional[bool]
        """
        Should return `True` or `None` in case the provided value can be converted.

        :param obj:
        :param field:
        :param value:
        :return:
        """
        pass

    def convert(self, obj, field, value):
        # type: (...) -> Any
        """
        Converts the provided `value`. This method is only called when `accepts()` has returned `True`.
        Implementors can dynamically declare that they are not able to convert the given value, by raising an Exception.

        Returning `None` means that the `value` converts to `None`.

        :param obj:
        :param field:
        :param value:
        :return:
        """
        raise NotImplementedError()

    @classmethod
    def create_from_fun(cls,
                        converter_fun,       # type: ConverterFuncOrLambda
                        validation_fun=None  # type: ValidationFuncOrLambda
                        ):
        # type: (...) -> Converter
        """
        Creates an instance of `Converter` where the `accepts` method is bound to the provided `validation_fun` and the
        `convert` method bound to the provided `converter_fun`.

        If these methods have less than 3 parameters, the mapping is done acccordingly.

        :param converter_fun:
        :param validation_fun:
        :return:
        """
        # Mandatory conversion callable
        if is_mini_lambda(converter_fun):
            is_mini = True
            converter_fun = converter_fun.as_function()
        else:
            is_mini = False
        converter_fun_3params = make_3params_callable(converter_fun, is_mini_lambda=is_mini)

        # Optional acceptance callable
        if validation_fun is not None:
            if is_mini_lambda(validation_fun):
                is_mini = True
                validation_fun = validation_fun.as_function()
            else:
                is_mini = False
            validation_fun_3params = make_3params_callable(validation_fun, is_mini_lambda=is_mini)
        else:
            validation_fun_3params = None

        # Finally create the converter instance
        return ConverterWithFuncs(name=converter_fun_3params.__name__,
                                  accepts_fun=validation_fun_3params,
                                  convert_fun=converter_fun_3params)


# noinspection PyAbstractClass
class ConverterWithFuncs(Converter):
    """
    Represents a converter for which the `accepts` and `convert` methods can be provided in the constructor.
    """
    __slots__ = ('accepts', 'convert')

    def __init__(self, convert_fun, name=None, accepts_fun=None):
        # call super to set the name
        super(ConverterWithFuncs, self).__init__(name=name)

        # use the convert method
        self.convert = convert_fun

        # use the accepts method if provided, otherwise use parent's
        if accepts_fun is not None:
            self.accepts = accepts_fun
        else:
            # use parent method - bind it on this instance
            self.accepts = Converter.accepts.__get__(self, ConverterWithFuncs)


if use_type_hints:
    # --------------converter type hints
    # 1. the lowest-level user or 3d party-provided validation functions
    ConverterFunc = Union[Callable[[Any], Any],
                          Callable[[Any, Any], Any],
                          Callable[[Any, Any, Any], Any]]
    """A converter function is a callable with signature (val), (obj, val) or (obj, field, val), returning the 
    converted value in case of success"""

    try:
        # noinspection PyUnresolvedReferences
        from mini_lambda import y

        ConverterFuncOrLambda = Union[ConverterFunc, type(y)]
    except ImportError:
        ConverterFuncOrLambda = ConverterFunc

    # 2. the syntax to optionally transform them into Converter by providing a tuple
    ValidType = Type
    # noinspection PyUnboundLocalVariable
    ConverterFuncDefinition = Union[Converter,
                                    ConverterFuncOrLambda,
                                    Tuple[ValidationFuncOrLambda, ConverterFuncOrLambda],
                                    Tuple[ValidType, ConverterFuncOrLambda]]

    TypeDef = Union[Type, Tuple[Type, ...], Literal['*'], str]  # todo remove str whe pycharm understands Literal
    OneOrSeveralConverterDefinitions = Union[Converter,
                                             ConverterFuncOrLambda,
                                             Iterable[Tuple[TypeDef, ConverterFuncOrLambda]],
                                             Mapping[TypeDef, ConverterFuncOrLambda]]
    Converters = OneOrSeveralConverterDefinitions


def make_3params_callable(f,                    # Union[ValidationFunc, ConverterFunc]
                          is_mini_lambda=False  # type: bool
                          ):
    # type: (...) -> Callable[[Any, 'Field', Any], Any]
    """
    Transforms the provided validation or conversion callable into a callable with 3 arguments (obj, field, val).

    :param f:
    :param is_mini_lambda: a boolean indicating if the function comes from a mini lambda. In which case we know the signature has one param only (x)
    :return:
    """
    # support several cases for the function signature
    # `f(val)`, `f(obj, val)` or `f(obj, field, val)`
    if is_mini_lambda:
        nbargs = 1
        nbvarargs = 0
        # nbkwargs = 0
        # nbdefaults = 0
    else:
        try:
            args, varargs, varkwargs, defaults = v8_getfullargspec(f, skip_bound_arg=True)[0:4]
            nbargs = len(args) if args is not None else 0
            nbvarargs = 1 if varargs is not None else 0
            # nbkwargs = 1 if varkwargs is not None else 0
            # nbdefaults = len(defaults) if defaults is not None else 0
        except IsBuiltInError:
            # built-ins: TypeError: <built-in function isinstance> is not a Python function
            # assume signature with a single positional argument
            nbargs = 1
            nbvarargs = 0
            # nbkwargs = 0
            # nbdefaults = 0

    if nbargs == 0 and nbvarargs == 0:
        raise ValueError(
            "validation or converter function should accept 1, 2, or 3 arguments at least. `f(val)`, `f(obj, val)` or "
            "`f(obj, field, val)`")
    elif nbargs == 1 or (
            nbargs == 0 and nbvarargs >= 1):  # varargs default to one argument (compliance with old mini lambda)
        # `f(val)`
        def new_f_with_3_args(obj, field, value):
            return f(value)

    elif nbargs == 2:
        # `f(obj, val)`
        def new_f_with_3_args(obj, field, value):
            return f(obj, value)

    else:
        # `f(obj, field, val, *opt_args, **ctx)`
        new_f_with_3_args = f

    # preserve the name
    new_f_with_3_args.__name__ = get_callable_name(f)

    return new_f_with_3_args


JOKER_STR = '*'
"""String used in converter definition dict entries or tuples, to indicate that the converter accepts everything"""


def make_converter(converter_def  # type: ConverterFuncDefinition
                   ):
    # type: (...) -> Converter
    """
    Makes a `Converter` from the provided converter object. Supported formats:

     - a `Converter`
     - a `<conversion_callable>` with possible signatures `(value)`, `(obj, value)`, `(obj, field, value)`.
     - a tuple `(<validation_callable>, <conversion_callable>)`
     - a tuple `(<valid_type>, <conversion_callable>)`

    If no name is provided and a `<conversion_callable>` is present, the callable name will be used as the converter
    name.

    The name of the conversion callable will be used in that case

    :param converter_def:
    :return:
    """
    try:
        nb_elts = len(converter_def)
    except (TypeError, FunctionDefinitionError):
        # -- single element
        # handle the special case of a LambdaExpression: automatically convert to a function
        if not is_mini_lambda(converter_def):
            if isinstance(converter_def, Converter):
                # already a converter
                return converter_def
            elif not callable(converter_def):
                raise ValueError('base converter function(s) not compliant with the allowed syntax. Base validation'
                                 ' function(s) can be %s Found %s.' % (supported_syntax, converter_def))
        # single element.
        return Converter.create_from_fun(converter_def)
    else:
        # -- a tuple
        if nb_elts == 1:
            converter_fun, validation_fun = converter_def[0], None
        elif nb_elts == 2:
            validation_fun, converter_fun = converter_def
            if validation_fun is not None:
                if isinstance(validation_fun, type):
                    # a type can be provided to denote accept "instances of <type>"
                    validation_fun = instance_of(validation_fun)
                elif validation_fun == JOKER_STR:
                    validation_fun = None
                else:
                    if not is_mini_lambda(validation_fun) and not callable(validation_fun):
                        raise ValueError('base converter function(s) not compliant with the allowed syntax. Validator '
                                         'is incorrect. Base converter function(s) can be %s Found %s.'
                                         % (supported_syntax, converter_def))
        else:
            raise ValueError(
                'tuple in converter_fun definition should have length 1, or 2. Found: %s' % (converter_def,))

        # check that the definition is valid
        if not is_mini_lambda(converter_fun) and not callable(converter_fun):
            raise ValueError('base converter function(s) not compliant with the allowed syntax. Base converter'
                             ' function(s) can be %s Found %s.' % (supported_syntax, converter_def))

        # finally create the failure raising callable
        return Converter.create_from_fun(converter_fun, validation_fun)


def make_converters_list(converters  # type: OneOrSeveralConverterDefinitions
                         ):
    # type: (...) -> Tuple[Converter, ...]
    """
    Creates a tuple of converters from the provided `converters`. The following things are supported:

     - a single item. This can be a `Converter`, a `<converter_callable>`, a tuple
       `(<acceptance_callable>, <converter_callable>)` or a tuple `(<accepted_type>, <converter_callable>)`.
       `<accepted_type>` can also contain `None` or `'*'`, both mean "anything".

     - a list of such items

     - a dictionary-like of `<acceptance>: <converter_callable>`, where `<acceptance>` can be an `<acceptance_callable>`
       or an `<accepted_type>`.

    :param converters:
    :return:
    """
    # support a single tuple
    if isinstance(converters, tuple):
        converters = [converters]

    try:
        # mapping ?
        c_items = iter(converters.items())
    except (AttributeError, FunctionDefinitionError):
        try:
            # iterable ?
            c_iter = iter(converters)
        except (TypeError, FunctionDefinitionError):
            # single converter: create a tuple manually
            all_converters = (make_converter(converters),)
        else:
            # iterable
            all_converters = tuple(make_converter(c) for c in c_iter)
    else:
        # mapping: assume that each entry is {validation_fun: converter_fun}
        all_converters = tuple(make_converter((k, v)) for k, v in c_items)

    if len(all_converters) == 0:
        raise ValueError("No converters provided")
    else:
        return all_converters


def trace_convert(field,    # type: 'Field'
                  value,    # type: Any
                  obj=None  # type: Any
                  ):
    # type: (...) -> Tuple[Any, DetailedConversionResults]
    """
    Utility method to debug conversion issues.
    Instead of just returning the converted value, it also returns conversion details.

    In case conversion can not be made, a `ConversionError` is raised.

    Inspired by the `getversion` library.

    :param obj:
    :param field:
    :param value:
    :return:
    """
    errors = OrderedDict()

    for conv in field.converters:
        try:
            # check if converter accepts this ?
            accepted = conv.accepts(obj, field, value)
        except Exception as e:
            # error in acceptance test
            errors[conv] = "Acceptance test: ERROR [%s] %s" % (e.__class__.__name__, e)
        else:
            if accepted is not None and not accepted:
                # acceptance failed
                errors[conv] = "Acceptance test: REJECTED (returned %s)" % accepted
            else:
                # accepted! (None or True truth value)
                try:
                    # apply converter
                    converted_value = conv.convert(obj, field, value)
                except Exception as e:
                    errors[conv] = "Acceptance test: SUCCESS (returned %s). Conversion: ERROR [%s] %s" \
                                   % (accepted, e.__class__.__name__, e)
                else:
                    # conversion success !
                    errors[conv] = "Acceptance test: SUCCESS (returned %s). Conversion: SUCCESS -> %s" \
                                   % (accepted, converted_value)
                    return converted_value, DetailedConversionResults(value, field, obj, errors, conv, converted_value)

    raise ConversionError(value_to_convert=value, field=field, obj=obj, err_dct=errors)


class ConversionError(Exception):
    """
    Final exception Raised by `trace_convert` when a value cannot be converted successfully
    """
    __slots__ = 'value_to_convert', 'field', 'obj', 'err_dct'

    def __init__(self, value_to_convert, field, obj, err_dct):
        self.value_to_convert = value_to_convert
        self.field = field
        self.obj = obj
        self.err_dct = err_dct
        super(ConversionError, self).__init__()

    def __str__(self):
        return "Unable to convert value %r. Results:\n%s" \
               % (self.value_to_convert, err_dct_to_str(self.err_dct))


def err_dct_to_str(err_dct  # Dict[Converter, str]
                   ):
    # type: (...) -> str
    msg = ""
    for converter, err in err_dct.items():
        msg += " - Converter '%s': %s\n" % (converter, err)

    return msg


class DetailedConversionResults(object):
    """
    Returned by `trace_convert` for detailed results about which converter failed before the winning one.
    """
    __slots__ = 'value_to_convert', 'field', 'obj', 'err_dct', 'winning_converter', 'converted_value'

    def __init__(self, value_to_convert, field, obj, err_dct, winning_converter, converted_value):
        self.value_to_convert= value_to_convert
        self.field = field
        self.obj = obj
        self.err_dct = err_dct
        self.winning_converter = winning_converter
        self.converted_value = converted_value

    def __str__(self):
        return "Value %r successfully converted to %r using converter '%s', after the following attempts:\n%s"\
               % (self.value_to_convert, self.converted_value, self.winning_converter, err_dct_to_str(self.err_dct))



from .typing_utils import FieldTypeError
from .core import field, Field, FieldError, MandatoryFieldInitError, UnsupportedOnNativeFieldError, \
    ReadOnlyFieldError, NoneError
from .validate_n_convert import Converter, ConversionError, DetailedConversionResults, trace_convert
from .init_makers import inject_fields, make_init, init_fields
from .helpers import copy_value, copy_field, copy_attr, has_fields, get_fields, yield_fields, get_field
from .autofields_ import autofields

try:
    # Distribution mode : import from _version.py generated by setuptools_scm during release
    from ._version import version as __version__
except ImportError:
    # Source mode : use setuptools_scm to get the current version from src using git
    from setuptools_scm import get_version as _gv
    from os import path as _path
    __version__ = _gv(_path.join(_path.dirname(__file__), _path.pardir))

__all__ = [
    '__version__',
    # submodules
    'core', 'validate_n_convert', 'init_makers', 'helpers', 'autofields_',
    # symbols
    'field', 'Field', 'FieldError', 'MandatoryFieldInitError', 'UnsupportedOnNativeFieldError',
    'ReadOnlyFieldError', 'FieldTypeError', 'NoneError',
    'Converter', 'ConversionError', 'DetailedConversionResults', 'trace_convert',
    'inject_fields', 'make_init', 'init_fields',
    'copy_value', 'copy_field', 'copy_attr', 'has_fields', 'get_fields', 'yield_fields', 'get_field',
    'autofields'
]



import os
import sys
import timeit

import pytest
from valid8 import ValidationError, ValidationFailure

from pyfields import field, MandatoryFieldInitError, make_init, init_fields, ReadOnlyFieldError, NoneError, \
    FieldTypeError


def runs_on_travis():
    return "TRAVIS_PYTHON_VERSION" in os.environ


def test_lazy_fields():

    class Wall(object):
        height = field(doc="Height of the wall in mm.")           # type: int
        color = field(default='white', doc="Color of the wall.")  # type: str

    # create an instance
    w = Wall()

    # the field is visible in `dir`
    assert dir(w)[-2:] == ['color', 'height']

    # but not yet in `vars`
    assert vars(w) == dict()

    # lets ask for it - default value is affected
    print(w.color)

    # now it is in `vars` too
    assert vars(w) == {'color': 'white'}

    # mandatory field
    with pytest.raises(MandatoryFieldInitError) as exc_info:
        print(w.height)
    assert str(exc_info.value).startswith("Mandatory field 'height' has not been initialized yet on instance <")

    w.height = 12
    assert vars(w) == {'color': 'white', 'height': 12}


@pytest.mark.parametrize("use_decorator", [False, True], ids="use_decorator={}".format)
def test_default_factory(use_decorator):

    class BadPocket(object):
        items = field(default=[])

    p = BadPocket()
    p.items.append('thing')
    g = BadPocket()
    assert g.items == ['thing']

    if use_decorator:
        class Pocket:
            items = field()

            @items.default_factory
            def default_items(self):
                return []
    else:
        class Pocket(object):
            items = field(default_factory=lambda obj: [])

    p = Pocket()
    g = Pocket()
    p.items.append('thing')
    assert p.items == ['thing']
    assert g.items == []


def test_readonly_field():
    """ checks that the example in the readme is correct """

    class User(object):
        name = field(read_only=True)

    u = User()
    u.name = "john"
    assert "name: %s" % u.name == "name: john"
    with pytest.raises(ReadOnlyFieldError) as exc_info:
        u.name = "john2"
    qualname = User.__dict__['name'].qualname
    assert str(exc_info.value) == "Read-only field '%s' has already been " \
                                  "initialized on instance %s and cannot be modified anymore." % (qualname, u)

    class User(object):
        name = field(read_only=True, default="dummy")

    u = User()
    assert "name: %s" % u.name == "name: dummy"
    with pytest.raises(ReadOnlyFieldError):
        u.name = "john"


@pytest.mark.parametrize("py36_style_type_hints", [False, True], ids="py36_style_type_hints={}".format)
def test_type_validation(py36_style_type_hints):
    if py36_style_type_hints:
        if sys.version_info < (3, 6):
            pytest.skip()
            Wall = None
        else:
            # import the test that uses python  3.6 type annotations
            from ._test_py36 import _test_readme_type_validation
            Wall = _test_readme_type_validation()
    else:
        class Wall(object):
            height = field(type_hint=int, check_type=True, doc="Height of the wall in mm.")
            color = field(type_hint=str, check_type=True, default='white', doc="Color of the wall.")

    w = Wall()
    w.height = 1
    with pytest.raises(TypeError):
        w.height = "1"


@pytest.mark.parametrize("py36_style_type_hints", [False, True], ids="py36_style_type_hints={}".format)
def test_value_validation(py36_style_type_hints):
    colors = ('blue', 'red', 'white')

    if py36_style_type_hints:
        if sys.version_info < (3, 6):
            pytest.skip()
            Wall = None
        else:
            # import the test that uses python  3.6 type annotations
            from ._test_py36 import _test_readme_value_validation
            Wall = _test_readme_value_validation(colors)

    from mini_lambda import x
    from valid8.validation_lib import is_in

    class Wall(object):
        height = field(type_hint=int,
                       validators={'should be a positive number': x > 0,
                                   'should be a multiple of 100': x % 100 == 0},
                       doc="Height of the wall in mm.")
        color = field(type_hint=str,
                      validators=is_in(colors),
                      default='white', doc="Color of the wall.")

    w = Wall()
    w.height = 100
    with pytest.raises(ValidationError) as exc_info:
        w.height = 1
    assert "Successes: ['x > 0'] / Failures: {" \
           "'x % 100 == 0': 'InvalidValue: should be a multiple of 100. Returned False.'" \
           "}." in str(exc_info.value)

    with pytest.raises(ValidationError) as exc_info:
        w.color = 'magenta'
    assert "NotInAllowedValues: x in ('blue', 'red', 'white') does not hold for x=magenta. Wrong value: 'magenta'." \
           in str(exc_info.value)


@pytest.mark.parametrize("py36_style_type_hints", [False, True], ids="py36_style_type_hints={}".format)
def test_value_validation_advanced(py36_style_type_hints):

    class InvalidWidth(ValidationFailure):
        help_msg = 'should be a multiple of the height ({height})'

    def validate_width(obj, width):
        if width % obj.height != 0:
            raise InvalidWidth(width, height=obj.height)

    if py36_style_type_hints:
        if sys.version_info < (3, 6):
            pytest.skip()
            Wall = None
        else:
            # import the test that uses python  3.6 type annotations
            from ._test_py36 import test_value_validation_advanced
            Wall = test_value_validation_advanced(validate_width)
    else:
        class Wall(object):
            height = field(type_hint=int,
                           doc="Height of the wall in mm.")
            width = field(type_hint=str,
                          validators=validate_width,
                          doc="Width of the wall in mm.")

    w = Wall()
    w.height = 100
    w.width = 200

    with pytest.raises(ValidationError) as exc_info:
        w.width = 201
    assert "InvalidWidth: should be a multiple of the height (100). Wrong value: 201." in str(exc_info.value)

try:
    from typing import Optional
    typing_present = True
except ImportError:
    typing_present = False


@pytest.mark.skipif(not typing_present, reason="typing module is not present")
@pytest.mark.parametrize("declaration", ['typing', 'default_value', 'explicit_nonable'], ids="declaration={}".format)
def test_nonable_fields(declaration):
    """Tests that nonable fields are supported and correctly handled"""

    if declaration == 'typing':
        from typing import Optional
        
        class Foo(object):
            a = field(type_hint=Optional[int], check_type=True)
            b = field(type_hint=Optional[int], validators={'is positive': lambda x: x > 0})
            c = field(nonable=False, check_type=True)
            d = field(validators={'accept_all': lambda x: True})
            e = field(nonable=False)

    elif declaration == 'default_value':
        class Foo(object):
            a = field(type_hint=int, default=None, check_type=True)
            b = field(type_hint=int, default=None, validators={'is positive': lambda x: x > 0})
            c = field(nonable=False, check_type=True)
            d = field(validators={'accept_all': lambda x: True})
            e = field(nonable=False)

    elif declaration == 'explicit_nonable':
        class Foo(object):
            a = field(type_hint=int, nonable=True, check_type=True)
            b = field(type_hint=int, nonable=True, validators={'is positive': lambda x: x > 0})
            c = field(nonable=False, check_type=True)
            d = field(validators={'accept_all': lambda x: True})
            e = field(nonable=False)

    else:
        raise ValueError(declaration)

    f = Foo()
    f.a = None
    f.b = None
    with pytest.raises(NoneError):
        f.c = None
    f.d = None
    f.e = None
    assert vars(f) == {'_a': None, '_b': None, '_d': None, 'e': None}


def test_native_descriptors():
    """"""
    class Foo:
        a = field()
        b = field(native=False)

    # TODO change when issue with class level access is fixed
    a_name = "test_native_descriptors.<locals>.Foo.a" if sys.version_info >= (3, 6) else "<unknown_cls>.None"
    b_name = "test_native_descriptors.<locals>.Foo.b" if sys.version_info >= (3, 6) else "<unknown_cls>.None"
    assert repr(Foo.__dict__['a']) == "<NativeField: %s>" % a_name
    assert repr(Foo.__dict__['b']) == "<DescriptorField: %s>" % b_name

    f = Foo()

    def set_a(): f.a = 12

    def set_b(): f.b = 12

    def set_c(): f.c = 12

    ta = timeit.Timer(set_a).timeit()
    tb = timeit.Timer(set_b).timeit()
    tc = timeit.Timer(set_c).timeit()

    print("Average time (ns) setting the field:")
    print("%0.2f (normal python) ; %0.2f (native field) ; %0.2f (descriptor field)" % (tc, ta, tb))

    print("Ratio is %.2f" % (ta / tc))

    # make sure that the access time for native field and native are identical
    # for reproducibility on travis, we have to get rid of the first init
    if runs_on_travis():
        print("increasing tolerance on travis.")
        assert ta / tc <= 2.0
    else:
        assert ta / tc <= 1.1
    # assert abs(round(t_field_native * 10) - round(t_native * 10)) <= 1


# def decompose(number):
#     """ decompose a number in scientific notation. from https://stackoverflow.com/a/45359185/7262247"""
#     (sign, digits, exponent) = Decimal(number).as_tuple()
#     fexp = len(digits) + exponent - 1
#     fman = Decimal(number).scaleb(-fexp).normalize()
#     return fman, fexp


def test_make_init_full_defaults():
    class Wall:
        height = field(doc="Height of the wall in mm.")           # type: int
        color = field(default='white', doc="Color of the wall.")  # type: str
        __init__ = make_init()

    # create an instance
    help(Wall)
    with pytest.raises(TypeError) as exc_info:
        Wall()
    assert str(exc_info.value).startswith("__init__()")

    w = Wall(2)
    assert vars(w) == {'color': 'white', 'height': 2}

    w = Wall(color='blue', height=12)
    assert vars(w) == {'color': 'blue', 'height': 12}


def test_make_init_with_explicit_list():
    class Wall:
        height = field(doc="Height of the wall in mm.")  # type: int
        color = field(default='white', doc="Color of the wall.")  # type: str

        # only `height` will be in the constructor
        __init__ = make_init(height)

    with pytest.raises(TypeError) as exc_info:
        Wall(1, 'blue')
    assert str(exc_info.value).startswith("__init__()")


def test_make_init_with_inheritance():
    class Wall:
        height = field(doc="Height of the wall in mm.")  # type: int
        __init__ = make_init(height)

    class ColoredWall(Wall):
        color = field(default='white', doc="Color of the wall.")  # type: str
        __init__ = make_init(Wall.__dict__['height'], color)

    w = ColoredWall(2)
    assert vars(w) == {'color': 'white', 'height': 2}

    w = ColoredWall(color='blue', height=12)
    assert vars(w) == {'color': 'blue', 'height': 12}


def test_make_init_callback():
    class Wall:
        height = field(doc="Height of the wall in mm.")  # type: int
        color = field(default='white', doc="Color of the wall.")  # type: str

        def post_init(self, msg='hello'):
            """
            After initialization, some print message is done
            :param msg: the message details to add
            :return:
            """
            print("post init ! height=%s, color=%s, msg=%s" % (self.height, self.color, msg))
            self.non_field_attr = msg

        # only `height` and `foo` will be in the constructor
        __init__ = make_init(height, post_init_fun=post_init)

    w = Wall(1, 'hey')
    assert vars(w) == {'color': 'white', 'height': 1, 'non_field_attr': 'hey'}


def test_init_fields():
    class Wall:
        height = field(doc="Height of the wall in mm.")  # type: int
        color = field(default='white', doc="Color of the wall.")  # type: str

        @init_fields
        def __init__(self, msg='hello'):
            """
            After initialization, some print message is done
            :param msg: the message details to add
            :return:
            """
            print("post init ! height=%s, color=%s, msg=%s" % (self.height, self.color, msg))
            self.non_field_attr = msg

    # create an instance
    help(Wall.__init__)
    with pytest.raises(TypeError) as exc_info:
        Wall()
    assert str(exc_info.value).startswith("__init__()")

    w = Wall(2)
    assert vars(w) == {'color': 'white', 'height': 2, 'non_field_attr': 'hello'}

    w = Wall(msg='hey', color='blue', height=12)
    assert vars(w) == {'color': 'blue', 'height': 12, 'non_field_attr': 'hey'}


no_type_checker = False
try:
    import typeguard
except ImportError:
    try:
        import pytypes
    except ImportError:
        no_type_checker = True


@pytest.mark.skipif(sys.version_info < (3, 6), reason="python < 3.6 does not support class member type hints")
@pytest.mark.skipif(no_type_checker, reason="no type checker is installed")
def test_autofields_readme():
    """Test for readme on autofields"""

    from ._test_py36 import _test_autofields_readme
    Pocket, Item, Pocket2 = _test_autofields_readme()

    with pytest.raises(TypeError):
        Item()

    item1 = Item(name='1')
    pocket1 = Pocket(size=2)
    pocket2 = Pocket(size=2)

    # make sure that custom constructor is not overridden by @autofields
    pocket3 = Pocket2("world")
    with pytest.raises(MandatoryFieldInitError):
        pocket3.size

    # make sure the items list is not the same in both (if we add the item to one, they do not appear in the 2d)
    assert pocket1.size == 2
    assert pocket1.items is not pocket2.items
    pocket1.items.append(item1)
    assert len(pocket2.items) == 0


try:
    import pytypes
except ImportError:
    has_pytypes = False
else:
    has_pytypes = True


@pytest.mark.skipif(has_pytypes, reason="pytypes does not correctly support vtypes - "
                                        "see https://github.com/Stewori/pytypes/issues/86")
@pytest.mark.skipif(sys.version_info < (3, 6), reason="python < 3.6 does not support class member type hints")
def test_autofields_vtypes_readme():

    from ._test_py36 import _test_autofields_vtypes_readme
    Rectangle = _test_autofields_vtypes_readme()

    r = Rectangle(1, 2)
    with pytest.raises(FieldTypeError):
        Rectangle(1, -2)
    with pytest.raises(FieldTypeError):
        Rectangle('1', 2)


def test_autoclass():
    """ Tests the example with autoclass in the doc """
    from autoclass import autoclass

    @autoclass
    class Foo(object):
        msg = field(type_hint=str)
        age = field(default=12, type_hint=int)

    foo = Foo(msg='hello')

    print(foo)  # automatic string representation
    print(dict(foo))  # dict view

    assert str(foo) == "Foo(msg='hello', age=12)"
    assert str(dict(foo)) in ("{'msg': 'hello', 'age': 12}", "{'age': 12, 'msg': 'hello'}")
    assert foo == Foo(msg='hello', age=12)  # comparison (equality)
    assert foo == {'msg': 'hello', 'age': 12}  # comparison with dicts



from typing import Union, Any, Callable, Iterable

from pyfields.core import Field


def init_fields(*fields: Union[Field, Any],
                init_args_before: bool = True,
                ancestor_fields_first: bool = True
                ):
    ...


def inject_fields(*fields: Union[Field, Any],
                  ):
    ...


def make_init(*fields: Union[Field, Any],
              post_init_fun: Callable = None,
              post_init_args_before: bool = True,
              ancestor_fields_first: bool = True
              ) -> Callable:
    ...


class InitDescriptor(object):
    def init(self, obj): ...
    ...


class InjectedInitFieldsArg(object):
    ...


def create_init(fields: Iterable[Field],
                user_init_fun: Callable[[...], Any] = None,
                inject_fields: bool = False,
                user_init_args_before: bool = True
                ):
    ...



#  Authors: Sylvain Marie <sylvain.marie@se.com>
#
#  Copyright (c) Schneider Electric Industries, 2019. All right reserved.
import sys

import pytest

from pyfields import field, init_fields, inject_fields, make_init, MandatoryFieldInitError, copy_field, get_fields
from pyfields.core import PY36


@pytest.mark.parametrize("native", [False, True], ids="native={}".format)
@pytest.mark.parametrize("init_type", ['inject_fields', 'make_init', 'make_init_with_postinit'],
                         ids="init_type={}".format)
@pytest.mark.parametrize("explicit_fields_list", [False, True], ids="explicit_list={}".format)
@pytest.mark.parametrize("py36_style_type_hints", [False, True], ids="py36_style_type_hints={}".format)
def test_init_all_methods(py36_style_type_hints, explicit_fields_list, init_type, native):
    """Test of @inject_fields with selected fields """
    if py36_style_type_hints:
        if sys.version_info < (3, 6):
            pytest.skip()
            Wall = None
        else:
            # import the test that uses python  3.6 type annotations
            from ._test_py36 import _test_readme_constructor
            Wall = _test_readme_constructor(explicit_fields_list, init_type, native)
    else:
        if init_type == 'inject_fields':
            if explicit_fields_list:
                class Wall(object):
                    height = field(doc="Height of the wall in mm.", native=native)          # type: int
                    color = field(default='white', doc="Color of the wall.", native=native)  # type: str

                    @inject_fields(height, color)
                    def __init__(self, fields):
                        with pytest.raises(MandatoryFieldInitError):
                            print(self.height)
                        # initialize all fields received
                        fields.init(self)
                        print(self.height)
            else:
                class Wall(object):
                    height = field(doc="Height of the wall in mm.", native=native)  # type: int
                    color = field(default='white', doc="Color of the wall.", native=native)  # type: str

                    @inject_fields
                    def __init__(self, fields):
                        with pytest.raises(MandatoryFieldInitError):
                            print(self.height)
                        # initialize all fields received
                        fields.init(self)
                        print(self.height)

        elif init_type == 'make_init':
            if explicit_fields_list:
                class Wall(object):
                    height = field(doc="Height of the wall in mm.", native=native)           # type: int
                    color = field(default='white', doc="Color of the wall.", native=native)  # type: str
                    __init__ = make_init(height, color)
            else:
                class Wall(object):
                    height = field(doc="Height of the wall in mm.", native=native)  # type: int
                    color = field(default='white', doc="Color of the wall.", native=native)  # type: str
                    __init__ = make_init()

        elif init_type == 'make_init_with_postinit':
            if explicit_fields_list:
                class Wall(object):
                    height = field(doc="Height of the wall in mm.", native=native)  # type: int
                    color = field(default='white', doc="Color of the wall.", native=native)  # type: str
                    def post_init(self, foo='bar'):
                        self.height
                        print("post init man !")

                    __init__ = make_init(height, color, post_init_fun=post_init)
            else:
                class Wall(object):
                    height = field(doc="Height of the wall in mm.", native=native)           # type: int
                    color = field(default='white', doc="Color of the wall.", native=native)  # type: str
                    def post_init(self, foo='bar'):
                        self.height
                        print("post init man !")

                    __init__ = make_init(post_init_fun=post_init)
        else:
            raise ValueError(init_type)

    # first init
    w = Wall(height=12)
    if native:
        assert vars(w) == {'color': 'white', 'height': 12}
    else:
        assert vars(w) == {'_color': 'white', '_height': 12}

    # make sure this can be done a second time (since we replaced the __init__ method now)
    w = Wall(color='blue', height=1)
    if native:
        assert vars(w) == {'color': 'blue', 'height': 1}
    else:
        assert vars(w) == {'_color': 'blue', '_height': 1}

    # type hints
    height_field = Wall.__dict__['height']
    color_field = Wall.__dict__['color']

    if py36_style_type_hints:
        assert height_field.type_hint is int
        assert color_field.type_hint is str

        # todo check signature of generated constructor


def test_init_partial_fields():
    class Wall(object):
        height = field(doc="Height of the wall in mm.")  # type: int
        color = field(default='white', doc="Color of the wall.")  # type: str

        def post_init(self, foo='bar'):
            print(self.height)
            print("post init man !")

        __init__ = make_init(height, post_init_fun=post_init)

    #
    # assert str(signature(Wall.__init__)) == "(self, height, foo='bar')"

    # first init
    w = Wall(12)
    assert vars(w) == {'height': 12}  # color not initialized yet

    # make sure this can be done a second time (since we replaced the __init__ method now)
    w = Wall(foo='blue', height=1)
    assert vars(w) == {'height': 1}  # color not initialized yet


def test_init_order():
    """Tests that order of initialization is the same than order of definition in the class"""

    class C(object):
        y = field()
        x = field(default_factory=copy_field(y))

        @init_fields
        def __init__(self):
            pass

    c = C(y=1)
    print(vars(c))


@pytest.mark.parametrize("a_first", [None, False, True], ids="ancestor_first={}".format)
def test_init_order2(a_first):
    """"""
    class A(object):
        a = field()
        d = field(default=5)

    class B(object):
        b = field()

    class C(B, A):
        a = field(default=None)
        c = field(default_factory=copy_field('b'))

        @init_fields(ancestor_fields_first=a_first)
        def __init__(self):
            pass

    fields = get_fields(C, include_inherited=True, ancestors_first=a_first if a_first is not None else True,
                        _auto_fix_fields=not PY36)
    field_names = [f.name for f in fields]
    if a_first is None or a_first:
        assert field_names == ['a', 'd', 'b', 'c']
    else:
        assert field_names == ['a', 'c', 'b', 'd']

    # make sure that a and c have default values and therefore just passing b is ok.
    c = C(1)
    assert vars(c) == {'b': 1, 'c': 1, 'a': None, 'd': 5}

    c = C(1, 2, 3)
    if a_first is None or a_first:
        assert vars(c) == {'b': 1,  # 1st arg
                           'c': 1,  # default: copy of b
                           'a': 2,  # 2d arg
                           'd': 3   # 3d arg
                           }
    else:
        assert vars(c) == {'b': 1,  # 1st arg
                           'c': 3,  # 3d arg
                           'a': 2,  # 2d arg
                           'd': 5   # default: 5
                           }


def test_init_inheritance():
    """Makes sure that the init method can be generated in inheritance case """

    class A(object):
        a = field(default='hello')

    class B(A):
        b = field(default='world')

        def a(self):
            """ purposedly override the base class field name """
            pass

        __init__ = make_init()

    # make sure that the 'a' field is ok
    b = B(a='h', b='w')
    assert b.a, b.b == ('h', 'w')



#  Authors: Sylvain Marie <sylvain.marie@se.com>
#
#  License: BSD 3 clause

from os.path import abspath

import click
from setuptools_scm import get_version


@click.command()
@click.argument('dest_folder')
def write_version(dest_folder):
    file_name = '%s/_version.py' % dest_folder
    print("Writing version to file: %s" % abspath(file_name))
    get_version('.', write_to=file_name)


if __name__ == '__main__':
    write_version()
